# Copyright (c) 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import six

from sahara import conductor as c
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.service.castellan import utils as key_manager
from sahara.service.edp import job_utils
from sahara.service import trusts as t
from sahara.swift import utils as su
from sahara.utils.openstack import base as b
from sahara.utils.openstack import keystone as k


PROXY_DOMAIN = None
conductor = c.API
LOG = logging.getLogger(__name__)
CONF = cfg.CONF

opts = [
    cfg.BoolOpt('use_domain_for_proxy_users',
                default=False,
                help='Enables Sahara to use a domain for creating temporary '
                     'proxy users to access Swift. If this is enabled '
                     'a domain must be created for Sahara to use.'),
    cfg.StrOpt('proxy_user_domain_name',
               default=None,
               help='The domain Sahara will use to create new proxy users '
                    'for Swift object access.'),
    cfg.ListOpt('proxy_user_role_names',
                default=['member'],
                help='A list of the role names that the proxy user should '
                     'assume through trust for Swift object access.')
]
CONF.register_opts(opts)


def create_proxy_user_for_job_execution(job_execution):
    '''Creates a proxy user and adds the credentials to the job execution

    :param job_execution: The job execution model to update

    '''
    username = 'job_{0}'.format(job_execution.id)
    password = key_manager.store_secret(proxy_user_create(username))
    current_user = k.auth()
    proxy_user = k.auth_for_proxy(username, password)
    trust_id = t.create_trust(trustor=current_user,
                              trustee=proxy_user,
                              role_names=CONF.proxy_user_role_names)
    update = {'job_configs': job_execution.job_configs.to_dict()}
    update['job_configs']['proxy_configs'] = {
        'proxy_username': username,
        'proxy_password': password,
        'proxy_trust_id': trust_id
        }
    conductor.job_execution_update(context.ctx(), job_execution, update)


def delete_proxy_user_for_job_execution(job_execution):
    '''Delete a proxy user based on a JobExecution

    :param job_execution: The job execution with proxy user information
    :returns: An updated job_configs dictionary or None

    '''
    proxy_configs = job_execution.job_configs.get('proxy_configs')
    if proxy_configs is not None:
        proxy_username = proxy_configs.get('proxy_username')
        proxy_trust_id = proxy_configs.get('proxy_trust_id')
        proxy_user = k.auth_for_proxy(proxy_username,
                                      key_manager.get_secret(
                                          proxy_configs.get('proxy_password')),
                                      proxy_trust_id)
        t.delete_trust(proxy_user, proxy_trust_id)
        proxy_user_delete(proxy_username)
        key_manager.delete_secret(proxy_configs.get('proxy_password'))
        update = job_execution.job_configs.to_dict()
        del update['proxy_configs']
        return update
    return None


def create_proxy_user_for_cluster(cluster):
    '''Creates a proxy user and adds the credentials to the cluster

    :param cluster: The cluster model to update

    '''
    if cluster.cluster_configs.get('proxy_configs'):
        return cluster
    username = 'cluster_{0}'.format(cluster.id)
    password = key_manager.store_secret(proxy_user_create(username))
    current_user = k.auth()
    proxy_user = k.auth_for_proxy(username, password)
    trust_id = t.create_trust(trustor=current_user,
                              trustee=proxy_user,
                              role_names=CONF.proxy_user_role_names)
    update = {'cluster_configs': cluster.cluster_configs.to_dict()}
    update['cluster_configs']['proxy_configs'] = {
        'proxy_username': username,
        'proxy_password': password,
        'proxy_trust_id': trust_id
        }
    return conductor.cluster_update(context.ctx(), cluster, update)


def delete_proxy_user_for_cluster(cluster):
    '''Delete a proxy user based on a Cluster

    :param cluster: The cluster model with proxy user information

    '''
    proxy_configs = cluster.cluster_configs.get('proxy_configs')
    if proxy_configs is not None:
        proxy_username = proxy_configs.get('proxy_username')
        proxy_trust_id = proxy_configs.get('proxy_trust_id')
        proxy_user = k.auth_for_proxy(proxy_username,
                                      key_manager.get_secret(
                                          proxy_configs.get('proxy_password')),
                                      proxy_trust_id)
        t.delete_trust(proxy_user, proxy_trust_id)
        proxy_user_delete(proxy_username)
        key_manager.delete_secret(proxy_configs.get('proxy_password'))
        update = {'cluster_configs': cluster.cluster_configs.to_dict()}
        del update['cluster_configs']['proxy_configs']
        conductor.cluster_update(context.ctx(), cluster, update)


def domain_for_proxy():
    '''Return the proxy domain or None

    If configured to use the proxy domain, this function will return that
    domain. If not configured to use the proxy domain, this function will
    return None. If the proxy domain can't be found this will raise an
    exception.

    :returns: A Keystone Domain object or None.
    :raises ConfigurationError: If the domain is requested but not specified.
    :raises NotFoundException: If the domain name is specified but cannot be
                               found.

    '''
    if CONF.use_domain_for_proxy_users is False:
        return None
    if CONF.proxy_user_domain_name is None:
        raise ex.ConfigurationError(_('Proxy domain requested but not '
                                      'specified.'))
    admin = k.client_for_admin()

    global PROXY_DOMAIN
    if not PROXY_DOMAIN:
        domain_list = b.execute_with_retries(
            admin.domains.list, name=CONF.proxy_user_domain_name)
        if len(domain_list) == 0:
            raise ex.NotFoundException(
                value=CONF.proxy_user_domain_name,
                message_template=_('Failed to find domain %s'))
        # the domain name should be globally unique in Keystone
        if len(domain_list) > 1:
            raise ex.NotFoundException(
                value=CONF.proxy_user_domain_name,
                message_template=_('Unexpected results found when searching '
                                   'for domain %s'))
        PROXY_DOMAIN = domain_list[0]
    return PROXY_DOMAIN


def job_execution_requires_proxy_user(job_execution):
    '''Returns True if the job execution requires a proxy user.'''

    def _check_values(values):
        return any(value.startswith(
            su.SWIFT_INTERNAL_PREFIX) for value in values if (
                isinstance(value, six.string_types)))

    if CONF.use_domain_for_proxy_users is False:
        return False

    paths = [conductor.data_source_get(context.ctx(), job_execution.output_id),
             conductor.data_source_get(context.ctx(), job_execution.input_id)]
    if _check_values(ds.url for ds in paths if ds):
        return True

    if _check_values(six.itervalues(
            job_execution.job_configs.get('configs', {}))):
        return True

    if _check_values(six.itervalues(
            job_execution.job_configs.get('params', {}))):
        return True

    if _check_values(job_execution.job_configs.get('args', [])):
        return True

    job = conductor.job_get(context.ctx(), job_execution.job_id)
    if _check_values(main.url for main in job.mains):
        return True

    if _check_values(lib.url for lib in job.libs):
        return True

    # We did the simple checks, now if data_source referencing is
    # enabled and we have values that could be a name or uuid,
    # query for data_sources that match and contain a swift path
    by_name, by_uuid = job_utils.may_contain_data_source_refs(
        job_execution.job_configs)
    if by_name:
        names = tuple(job_utils.find_possible_data_source_refs_by_name(
            job_execution.job_configs))
        # do a query here for name in names and path starts with swift-prefix
        if names and conductor.data_source_count(
                context.ctx(),
                name=names,
                url=su.SWIFT_INTERNAL_PREFIX+'%') > 0:
            return True

    if by_uuid:
        uuids = tuple(job_utils.find_possible_data_source_refs_by_uuid(
            job_execution.job_configs))
        # do a query here for id in uuids and path starts with swift-prefix
        if uuids and conductor.data_source_count(
                context.ctx(),
                id=uuids,
                url=su.SWIFT_INTERNAL_PREFIX+'%') > 0:
            return True

    return False


def proxy_domain_users_list():
    '''Return a list of all users in the proxy domain.'''
    admin = k.client_for_admin()
    domain = domain_for_proxy()
    if domain:
        return b.execute_with_retries(admin.users.list, domain=domain.id)
    return []


def proxy_user_create(username):
    '''Create a new user in the proxy domain

    Creates the username specified with a random password.

    :param username: The name of the new user.
    :returns: The password created for the user.

    '''
    admin = k.client_for_admin()
    domain = domain_for_proxy()
    password = uuidutils.generate_uuid()
    b.execute_with_retries(
        admin.users.create, name=username, password=password, domain=domain.id)
    LOG.debug('Created proxy user {username}'.format(username=username))
    return password


def proxy_user_delete(username=None, user_id=None):
    '''Delete the user from the proxy domain.

    :param username: The name of the user to delete.
    :param user_id: The id of the user to delete, if provided this overrides
                    the username.
    :raises NotFoundException: If there is an error locating the user in the
                               proxy domain.

    '''
    admin = k.client_for_admin()
    if not user_id:
        domain = domain_for_proxy()
        user_list = b.execute_with_retries(
            admin.users.list, domain=domain.id, name=username)
        if len(user_list) == 0:
            raise ex.NotFoundException(
                value=username,
                message_template=_('Failed to find user %s'))
        if len(user_list) > 1:
            raise ex.NotFoundException(
                value=username,
                message_template=_('Unexpected results found when searching '
                                   'for user %s'))
        user_id = user_list[0].id
    b.execute_with_retries(admin.users.delete, user_id)
    LOG.debug('Deleted proxy user id {user_id}'.format(user_id=user_id))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

# loading keystonemiddleware opts because sahara uses these options in code
from keystonemiddleware import opts  # noqa
from oslo_config import cfg
from oslo_log import log

from sahara import exceptions as ex
from sahara.i18n import _
from sahara.plugins import opts as plugins_base
from sahara.service.castellan import config as castellan
from sahara.service.edp.data_sources import opts as data_source
from sahara.service.edp.job_binaries import opts as job_binary
from sahara.topology import topology_helper
from sahara.utils.notification import sender
from sahara.utils.openstack import cinder
from sahara.utils.openstack import keystone
from sahara.utils import remote
from sahara import version


cli_opts = [
    cfg.HostAddressOpt('host', default='0.0.0.0',
                       help='Hostname or IP address that will be used '
                            'to listen on.'),
    cfg.PortOpt('port', default=8386,
                help='Port that will be used to listen on.'),
    cfg.BoolOpt('log-exchange', default=False,
                help='Log request/response exchange details: environ, '
                     'headers and bodies.')
]

edp_opts = [
    cfg.IntOpt('job_binary_max_KB',
               default=5120,
               help='Maximum length of job binary data in kilobytes that '
                    'may be stored or retrieved in a single operation.'),
    cfg.IntOpt('job_canceling_timeout',
               default=300,
               help='Timeout for canceling job execution (in seconds). '
                    'Sahara will try to cancel job execution during '
                    'this time.'),
    cfg.BoolOpt('edp_internal_db_enabled',
                default=True,
                help='Use Sahara internal db to store job binaries.')
]

db_opts = [
    cfg.StrOpt('db_driver',
               default='sahara.db',
               help='Driver to use for database access.')
]

networking_opts = [
    cfg.BoolOpt('use_floating_ips',
                default=True,
                help='If set to True, Sahara will use floating IPs to '
                     'communicate with instances. To make sure that all '
                     'instances have floating IPs assigned, make sure '
                     'that all Node Groups have "floating_ip_pool" '
                     'parameter defined.'),
    cfg.StrOpt('node_domain',
               default='novalocal',
               help="The suffix of the node's FQDN."),
    cfg.BoolOpt('use_namespaces',
                default=False,
                help="Use network namespaces for communication."),
    cfg.BoolOpt('use_rootwrap',
                default=False,
                help="Use rootwrap facility to allow non-root users to run "
                     "the sahara services and access private network IPs "
                     "(only valid to use in conjunction with "
                     "use_namespaces=True)"),
    cfg.StrOpt('rootwrap_command',
               default='sudo sahara-rootwrap /etc/sahara/rootwrap.conf',
               help="Rootwrap command to leverage.  Use in conjunction with "
                    "use_rootwrap=True")
]

dns_opts = [
    cfg.BoolOpt('use_designate',
                default=False,
                help='Use Designate for internal and external hostnames '
                     'resolution'),
    cfg.ListOpt('nameservers',
                default=[],
                help="IP addresses of Designate nameservers. "
                     "This is required if 'use_designate' is True")
]

accessible_ip_opts = [
    cfg.IPOpt('identity_ip_accessible',
              default=None,
              help='IP address of Keystone endpoint, accessible by tenant'
                   ' machines. If not set, the results of the DNS lookup'
                   ' performed where Sahara services are running will be'
                   ' used.'),
    cfg.IPOpt('object_store_ip_accessible',
              default=None,
              help='IP address of Swift endpoint, accessible by tenant'
                   ' machines. If not set, the results of the DNS lookup'
                   ' performed where Sahara services are running will be'
                   ' used.'),
]

CONF = cfg.CONF
CONF.register_cli_opts(cli_opts)
CONF.register_opts(networking_opts)
CONF.register_opts(edp_opts)
CONF.register_opts(db_opts)
CONF.register_opts(dns_opts)
CONF.register_opts(accessible_ip_opts)

log.register_options(CONF)

sahara_default_log_levels = [
    'stevedore=INFO',
    'eventlet.wsgi.server=WARN',
    'paramiko=WARN',
    'requests=WARN',
    'neutronclient=INFO',
]

log.set_defaults(
    default_log_levels=log.get_default_log_levels()+sahara_default_log_levels)


def list_opts():
    # NOTE (vgridnev): we make these import here to avoid problems
    #                  with importing unregistered options in sahara code.
    #                  As example, importing 'node_domain' in
    #                  sahara/conductor/objects.py

    from sahara.conductor import api
    from sahara import main as sahara_main
    from sahara.service import coordinator
    from sahara.service.edp import job_utils
    from sahara.service.heat import heat_engine
    from sahara.service.heat import templates
    from sahara.service import ntp_service
    from sahara.service import periodic
    from sahara.swift import swift_helper
    from sahara.utils import cluster_progress_ops as cpo
    from sahara.utils.openstack import base
    from sahara.utils.openstack import glance
    from sahara.utils.openstack import heat
    from sahara.utils.openstack import manila
    from sahara.utils.openstack import neutron
    from sahara.utils.openstack import nova
    from sahara.utils.openstack import swift
    from sahara.utils import poll_utils
    from sahara.utils import proxy
    from sahara.utils import ssh_remote

    return [
        (None,
         itertools.chain(cli_opts,
                         edp_opts,
                         networking_opts,
                         dns_opts,
                         db_opts,
                         accessible_ip_opts,
                         plugins_base.opts,
                         topology_helper.opts,
                         keystone.opts,
                         remote.ssh_opts,
                         sahara_main.opts,
                         job_utils.opts,
                         periodic.periodic_opts,
                         coordinator.coordinator_opts,
                         ntp_service.ntp_opts,
                         proxy.opts,
                         cpo.event_log_opts,
                         base.opts,
                         heat_engine.heat_engine_opts,
                         templates.heat_engine_opts,
                         ssh_remote.ssh_config_options,
                         castellan.opts,
                         data_source.opts,
                         job_binary.opts)),
        (poll_utils.timeouts.name,
         itertools.chain(poll_utils.timeouts_opts)),
        (api.conductor_group.name,
         itertools.chain(api.conductor_opts)),
        (cinder.cinder_group.name,
         itertools.chain(cinder.opts)),
        (glance.glance_group.name,
         itertools.chain(glance.opts)),
        (heat.heat_group.name,
         itertools.chain(heat.opts)),
        (manila.manila_group.name,
         itertools.chain(manila.opts)),
        (neutron.neutron_group.name,
         itertools.chain(neutron.opts)),
        (nova.nova_group.name,
         itertools.chain(nova.opts)),
        (swift.swift_group.name,
         itertools.chain(swift.opts)),
        (keystone.keystone_group.name,
         itertools.chain(keystone.ssl_opts)),
        (keystone.trustee_group.name,
         itertools.chain(keystone.trustee_opts)),
        (base.retries.name,
         itertools.chain(base.opts)),
        (swift_helper.public_endpoint_cert_group.name,
         itertools.chain(swift_helper.opts)),
        (castellan.castellan_group.name,
         itertools.chain(castellan.castellan_opts)),
        (sender.notifier_opts_group, sender.notifier_opts)
    ]


def parse_configs(conf_files=None):
    try:
        version_string = version.version_info.version_string()
        CONF(project='sahara', version=version_string,
             default_config_files=conf_files)
    except cfg.RequiredOptError as roe:
        raise ex.ConfigurationError(
            _("Option '%(option)s' is required for config group '%(group)s'") %
            {'option': roe.opt_name, 'group': roe.group.name})



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from sahara import context
from sahara.plugins import swift_utils as su
from sahara.utils import xmlutils as x


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
HADOOP_SWIFT_AUTH_URL = 'fs.swift.service.sahara.auth.url'
HADOOP_SWIFT_TENANT = 'fs.swift.service.sahara.tenant'
HADOOP_SWIFT_USERNAME = 'fs.swift.service.sahara.username'
HADOOP_SWIFT_PASSWORD = 'fs.swift.service.sahara.password'
HADOOP_SWIFT_REGION = 'fs.swift.service.sahara.region'
HADOOP_SWIFT_TRUST_ID = 'fs.swift.service.sahara.trust.id'
HADOOP_SWIFT_DOMAIN_NAME = 'fs.swift.service.sahara.domain.name'

opts = [
    cfg.StrOpt("public_identity_ca_file",
               help=("Location of ca certificate file to use for identity "
                     "client requests via public endpoint")),
    cfg.StrOpt("public_object_store_ca_file",
               help=("Location of ca certificate file to use for object-store "
                     "client requests via public endpoint"))
]

public_endpoint_cert_group = cfg.OptGroup(
    name="object_store_access", title="Auth options for Swift access from VM")

CONF.register_group(public_endpoint_cert_group)
CONF.register_opts(opts, group=public_endpoint_cert_group)


def retrieve_tenant():
    return context.current().tenant_name


def get_swift_configs():
    configs = x.load_hadoop_xml_defaults('swift/resources/conf-template.xml')
    for conf in configs:
        if conf['name'] == HADOOP_SWIFT_AUTH_URL:
            conf['value'] = su.retrieve_auth_url() + "auth/tokens/"
        if conf['name'] == HADOOP_SWIFT_TENANT:
            conf['value'] = retrieve_tenant()
        if CONF.os_region_name and conf['name'] == HADOOP_SWIFT_REGION:
            conf['value'] = CONF.os_region_name
        if conf['name'] == HADOOP_SWIFT_DOMAIN_NAME:
            # NOTE(jfreud): Don't be deceived here... Even though there is an
            # attribute provided by context called domain_name, it is used for
            # domain scope, and hadoop-swiftfs always authenticates using
            # project scope. The purpose of the setting below is to override
            # the default value for project domain and user domain, domain id
            # as 'default', which may not always be correct.
            # TODO(jfreud): When hadoop-swiftfs allows it, stop hoping that
            # project_domain_name is always equal to user_domain_name.
            conf['value'] = context.current().project_domain_name

    result = [cfg for cfg in configs if cfg['value']]
    LOG.info("Swift would be integrated with the following "
             "params: {result}".format(result=result))
    return result


def read_default_swift_configs():
    return x.load_hadoop_xml_defaults('swift/resources/conf-template.xml')


def install_ssl_certs(instances):
    certs = []
    if CONF.object_store_access.public_identity_ca_file:
        certs.append(CONF.object_store_access.public_identity_ca_file)
    if CONF.object_store_access.public_object_store_ca_file:
        certs.append(CONF.object_store_access.public_object_store_ca_file)
    if not certs:
        return
    with context.ThreadGroup() as tg:
        for inst in instances:
            tg.spawn("configure-ssl-cert-%s" % inst.instance_id,
                     _install_ssl_certs, inst, certs)


def _install_ssl_certs(instance, certs):
    register_cmd = (
        "sudo su - -c \"keytool -import -alias sahara-%d -keystore "
        "`cut -f2 -d \\\"=\\\" /etc/profile.d/99-java.sh | head -1`"
        "/lib/security/cacerts -file /tmp/cert.pem -noprompt -storepass "
        "changeit\"")
    with instance.remote() as r:
        for idx, cert in enumerate(certs):
            with open(cert) as cert_fd:
                data = cert_fd.read()
            r.write_file_to("/tmp/cert.pem", data)
            try:
                r.execute_command(register_cmd % idx)
            finally:
                r.execute_command("rm /tmp/cert.pem")



# Copyright 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""set is_protected on is_default

Revision ID: 029
Revises: 028
Create Date: 2015-11-4 12:41:52.571258

"""

# revision identifiers, used by Alembic.
revision = '029'
down_revision = '028'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


def upgrade():
    ng = table('node_group_templates',
               column('is_protected', sa.Boolean),
               column('is_default', sa.Boolean))
    op.execute(
        ng.update().where(
            ng.c.is_default).values({'is_protected': True})
    )

    clt = table('cluster_templates',
                column('is_protected', sa.Boolean),
                column('is_default', sa.Boolean))
    op.execute(
        clt.update().where(
            clt.c.is_default).values({'is_protected': True})
    )



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Icehouse release

Revision ID: 001
Revises: None
Create Date: 2014-04-01 20:46:25.783444

"""

# revision identifiers, used by Alembic.
revision = '001'
down_revision = None

from alembic import op
import sqlalchemy as sa

from sahara.db.sqlalchemy import types as st

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    op.create_table('jobs',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('type', sa.String(length=80), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('node_group_templates',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('flavor_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('image_id', sa.String(length=36), nullable=True),
                    sa.Column('plugin_name', sa.String(length=80),
                              nullable=False),
                    sa.Column('hadoop_version', sa.String(length=80),
                              nullable=False),
                    sa.Column('node_processes', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('node_configs', st.JsonEncoded(), nullable=True),
                    sa.Column('volumes_per_node', sa.Integer(),
                              nullable=False),
                    sa.Column('volumes_size', sa.Integer(), nullable=True),
                    sa.Column('volume_mount_prefix', sa.String(length=80),
                              nullable=True),
                    sa.Column('floating_ip_pool', sa.String(length=36),
                              nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('cluster_templates',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('cluster_configs', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('default_image_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('anti_affinity', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('neutron_management_network',
                              sa.String(length=36), nullable=True),
                    sa.Column('plugin_name', sa.String(length=80),
                              nullable=False),
                    sa.Column('hadoop_version', sa.String(length=80),
                              nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('job_binary_internal',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('data', st.LargeBinary(), nullable=True),
                    sa.Column('datasize', sa.BIGINT(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('job_binaries',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('url', sa.String(length=256), nullable=False),
                    sa.Column('extra', st.JsonEncoded(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('data_sources',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('type', sa.String(length=80), nullable=False),
                    sa.Column('url', sa.String(length=256), nullable=False),
                    sa.Column('credentials', st.JsonEncoded(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('libs_association',
                    sa.Column('Job_id', sa.String(length=36), nullable=True),
                    sa.Column('JobBinary_id', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['JobBinary_id'],
                                            ['job_binaries.id'], ),
                    sa.ForeignKeyConstraint(['Job_id'], ['jobs.id'], ),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('clusters',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('description', sa.Text(), nullable=True),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('trust_id', sa.String(length=36), nullable=True),
                    sa.Column('is_transient', sa.Boolean(), nullable=True),
                    sa.Column('plugin_name', sa.String(length=80),
                              nullable=False),
                    sa.Column('hadoop_version', sa.String(length=80),
                              nullable=False),
                    sa.Column('cluster_configs', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('default_image_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('neutron_management_network',
                              sa.String(length=36), nullable=True),
                    sa.Column('anti_affinity', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('management_private_key', sa.Text(),
                              nullable=False),
                    sa.Column('management_public_key', sa.Text(),
                              nullable=False),
                    sa.Column('user_keypair_id', sa.String(length=80),
                              nullable=True),
                    sa.Column('status', sa.String(length=80), nullable=True),
                    sa.Column('status_description', sa.String(length=200),
                              nullable=True),
                    sa.Column('info', st.JsonEncoded(), nullable=True),
                    sa.Column('extra', st.JsonEncoded(), nullable=True),
                    sa.Column('cluster_template_id', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['cluster_template_id'],
                                            ['cluster_templates.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'tenant_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('templates_relations',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('flavor_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('image_id', sa.String(length=36), nullable=True),
                    sa.Column('node_processes', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('node_configs', st.JsonEncoded(), nullable=True),
                    sa.Column('volumes_per_node', sa.Integer(), nullable=True),
                    sa.Column('volumes_size', sa.Integer(), nullable=True),
                    sa.Column('volume_mount_prefix', sa.String(length=80),
                              nullable=True),
                    sa.Column('count', sa.Integer(), nullable=False),
                    sa.Column('cluster_template_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('node_group_template_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('floating_ip_pool', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['cluster_template_id'],
                                            ['cluster_templates.id'], ),
                    sa.ForeignKeyConstraint(['node_group_template_id'],
                                            ['node_group_templates.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('mains_association',
                    sa.Column('Job_id', sa.String(length=36), nullable=True),
                    sa.Column('JobBinary_id', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['JobBinary_id'],
                                            ['job_binaries.id'], ),
                    sa.ForeignKeyConstraint(['Job_id'], ['jobs.id'], ),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('job_executions',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('job_id', sa.String(length=36), nullable=True),
                    sa.Column('input_id', sa.String(length=36), nullable=True),
                    sa.Column('output_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('start_time', sa.DateTime(), nullable=True),
                    sa.Column('end_time', sa.DateTime(), nullable=True),
                    sa.Column('cluster_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('info', st.JsonEncoded(), nullable=True),
                    sa.Column('progress', sa.Float(), nullable=True),
                    sa.Column('oozie_job_id', sa.String(length=100),
                              nullable=True),
                    sa.Column('return_code', sa.String(length=80),
                              nullable=True),
                    sa.Column('job_configs', st.JsonEncoded(), nullable=True),
                    sa.Column('extra', st.JsonEncoded(), nullable=True),
                    sa.ForeignKeyConstraint(['cluster_id'], ['clusters.id'], ),
                    sa.ForeignKeyConstraint(['input_id'],
                                            ['data_sources.id'], ),
                    sa.ForeignKeyConstraint(['job_id'], ['jobs.id'], ),
                    sa.ForeignKeyConstraint(['output_id'],
                                            ['data_sources.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('node_groups',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('name', sa.String(length=80), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('flavor_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('image_id', sa.String(length=36), nullable=True),
                    sa.Column('image_username', sa.String(length=36),
                              nullable=True),
                    sa.Column('node_processes', st.JsonEncoded(),
                              nullable=True),
                    sa.Column('node_configs', st.JsonEncoded(), nullable=True),
                    sa.Column('volumes_per_node', sa.Integer(), nullable=True),
                    sa.Column('volumes_size', sa.Integer(), nullable=True),
                    sa.Column('volume_mount_prefix', sa.String(length=80),
                              nullable=True),
                    sa.Column('count', sa.Integer(), nullable=False),
                    sa.Column('cluster_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('node_group_template_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('floating_ip_pool', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['cluster_id'], ['clusters.id'], ),
                    sa.ForeignKeyConstraint(['node_group_template_id'],
                                            ['node_group_templates.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'cluster_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)

    op.create_table('instances',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('node_group_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('instance_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('instance_name', sa.String(length=80),
                              nullable=False),
                    sa.Column('internal_ip', sa.String(length=15),
                              nullable=True),
                    sa.Column('management_ip', sa.String(length=15),
                              nullable=True),
                    sa.Column('volumes', st.JsonEncoded(), nullable=True),
                    sa.ForeignKeyConstraint(['node_group_id'],
                                            ['node_groups.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('instance_id', 'node_group_id'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import six


NATURAL_SORT_RE = re.compile('([0-9]+)')


def find_dict(iterable, **rules):
    """Search for dict in iterable of dicts using specified key-value rules."""

    for item in iterable:
        # assert all key-value pairs from rules dict
        ok = True
        for k, v in six.iteritems(rules):
            ok = ok and k in item and item[k] == v

        if ok:
            return item

    return None


def find(lst, **kwargs):
    for obj in lst:
        match = True
        for attr, value in kwargs.items():
            if getattr(obj, attr) != value:
                match = False

        if match:
            return obj

    return None


def get_by_id(lst, id):
    for obj in lst:
        if obj.id == id:
            return obj

    return None


# Taken from http://stackoverflow.com/questions/4836710/does-
# python-have-a-built-in-function-for-string-natural-sort
def natural_sort_key(s):
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split(NATURAL_SORT_RE, s)]


def generate_instance_name(cluster_name, node_group_name, index):
    return ("%s-%s-%03d" % (cluster_name, node_group_name, index)).lower()


def generate_auto_security_group_name(node_group):
    return ("%s-%s-%s" % (node_group.cluster.name, node_group.name,
                          node_group.id[:8])).lower()


def generate_aa_group_name(cluster_name, server_group_index):
    return ("%s-aa-group-%d" % (cluster_name, server_group_index)).lower()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy

import sahara.service.validations.edp.base as b


JOB_EXEC_SCHEMA = {
    "type": "object",
    "properties": {
        "input_id": {
            "type": "string",
            "format": "uuid",
        },
        "output_id": {
            "type": "string",
            "format": "uuid",
        },
        "cluster_id": {
            "type": "string",
            "format": "uuid",
        },
        "interface": {
            "type": "simple_config",
        },
        "job_configs": b.job_configs,
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        }
    },
    "additionalProperties": False,
    "required": [
        "cluster_id"
    ]
}


JOB_EXEC_SCHEMA_V2 = copy.deepcopy(JOB_EXEC_SCHEMA)
JOB_EXEC_SCHEMA_V2['properties'].update({
    "job_template_id": {
        "type": "string",
        "format": "uuid",
    }})
JOB_EXEC_SCHEMA_V2['required'].append('job_template_id')


JOB_EXEC_UPDATE_SCHEMA = {
    "type": "object",
    "properties": {
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        },
        "info": {
            "type": "simple_config",
            "properties": {
                    "status": {
                        "enum": ["suspend", "cancel"]
                    }
            },
            "additionalProperties": False
        }
    },
    "additionalProperties": False,
    "required": []
}



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from keystoneclient.v2_0 import Client as keystone_client
from oslo_config import cfg


possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                                os.pardir,
                                                os.pardir))
if os.path.exists(os.path.join(possible_topdir,
                               'sahara',
                               '__init__.py')):
    sys.path.insert(0, possible_topdir)

cli_opts = [
    cfg.StrOpt('username', default='',
               help='set username'),
    cfg.StrOpt('password', default='',
               help='set password'),
    cfg.StrOpt('tenant', default='',
               help='set tenant'),
]

CONF = cfg.CONF
CONF.import_opt('auth_uri', 'keystonemiddleware.auth_token', group='keystone_authtoken')
CONF.import_opt('admin_user', 'keystonemiddleware.auth_token', group='keystone_authtoken')
CONF.import_opt('admin_password', 'keystonemiddleware.auth_token', group='keystone_authtoken')
CONF.import_opt('admin_tenant_name', 'keystonemiddleware.auth_token', group='keystone_authtoken')
CONF.register_cli_opts(cli_opts)


def main():
    dev_conf = os.path.join(possible_topdir,
                            'etc',
                            'sahara',
                            'sahara.conf')
    config_files = None
    if os.path.exists(dev_conf):
        config_files = [dev_conf]

    CONF(sys.argv[1:], project='get_auth_token',
         default_config_files=config_files)

    auth_uri = CONF.keystone_authtoken.auth_uri
    user = CONF.username or CONF.keystone_authtoken.admin_user
    password = CONF.password or CONF.keystone_authtoken.admin_password
    tenant = CONF.tenant or CONF.keystone_authtoken.admin_tenant_name

    print "User: %s" % user
    print "Password: %s" % password
    print "Tenant: %s" % tenant
    print "Auth URI: %s" % auth_uri

    keystone = keystone_client(
        username=user,
        password=password,
        tenant_name=tenant,
        auth_url=auth_uri
    )

    result = keystone.authenticate()

    print "Auth succeed: %s" % result
    print "Auth token: %s" % keystone.auth_token
    print "Tenant [%s] id: %s" % (tenant, keystone.tenant_id)
    print "For bash:"
    print "export TOKEN=%s" % keystone.auth_token
    print "export TENANT_ID=%s" % keystone.tenant_id


if __name__ == "__main__":
    main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from sahara.service.validations import shares

NODE_GROUP_TEMPLATE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name_hostname",
        },
        "flavor_id": {
            "type": "flavor",
        },
        "plugin_name": {
            "type": "string",
        },
        "hadoop_version": {
            "type": "string",
        },
        "node_processes": {
            "type": "array",
            "items": {
                "type": "string",
            },
            "minItems": 1
        },
        "image_id": {
            "type": ["string", "null"],
            "format": "uuid",
        },
        "node_configs": {
            "type": ["configs", "null"],
        },
        "volumes_per_node": {
            "type": "integer",
            "minimum": 0,
        },
        "volumes_size": {
            "type": ["integer", "null"],
            "minimum": 1,
        },
        "volume_type": {
            "type": ["string", "null"],
        },
        "volumes_availability_zone": {
            "type": ["string", "null"],
        },
        "volume_mount_prefix": {
            "type": ["string", "null"],
            "format": "posix_path",
        },
        "description": {
            "type": ["string", "null"],
        },
        "floating_ip_pool": {
            "type": ["string", "null"],
        },
        "security_groups": {
            "type": ["array", "null"],
            "items": {"type": "string"}
        },
        "auto_security_group": {
            "type": ["boolean", "null"],
        },
        "availability_zone": {
            "type": ["string", "null"],
        },
        "is_proxy_gateway": {
            "type": ["boolean", "null"],
        },
        "volume_local_to_instance": {
            "type": ["boolean", "null"]
        },
        "shares": copy.deepcopy(shares.SHARE_SCHEMA),
        "use_autoconfig": {
            "type": ["boolean", "null"]
        },
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        }
    },
    "additionalProperties": False,
    "required": [
        "name",
        "flavor_id",
        "plugin_name",
        "hadoop_version",
        "node_processes",
    ]
}

# APIv2: renaming hadoop_version -> plugin_version
NODE_GROUP_TEMPLATE_SCHEMA_V2 = copy.deepcopy(NODE_GROUP_TEMPLATE_SCHEMA)
del NODE_GROUP_TEMPLATE_SCHEMA_V2["properties"]["hadoop_version"]
NODE_GROUP_TEMPLATE_SCHEMA_V2["required"].remove("hadoop_version")
NODE_GROUP_TEMPLATE_SCHEMA_V2["properties"].update({
    "plugin_version": {
        "type": "string",
    }})
NODE_GROUP_TEMPLATE_SCHEMA_V2["required"].append("plugin_version")
NODE_GROUP_TEMPLATE_SCHEMA_V2["properties"].update({
    "boot_from_volume": {
        "type": "boolean",
    },
    "boot_volume_type": {
        "type": "string",
    },
    "boot_volume_availability_zone": {
        "type": "string",
    },
    "boot_volume_local_to_instance": {
        "type": "boolean",
    }})


# For an update we do not require any fields but we want the given
# fields to be validated
NODE_GROUP_TEMPLATE_UPDATE_SCHEMA = copy.copy(NODE_GROUP_TEMPLATE_SCHEMA)
NODE_GROUP_TEMPLATE_UPDATE_SCHEMA["required"] = []

NODE_GROUP_TEMPLATE_UPDATE_SCHEMA_V2 = copy.copy(NODE_GROUP_TEMPLATE_SCHEMA_V2)
NODE_GROUP_TEMPLATE_UPDATE_SCHEMA_V2["required"] = []



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Mirantis Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Tests for database migrations.

For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost.
The test will then use that db and u/p combo to run the tests.

For postgres on Ubuntu this can be done with the following commands:

sudo -u postgres psql
postgres=# create user openstack_citest with createdb login password
      'openstack_citest';
postgres=# create database openstack_citest with owner openstack_citest;

"""

import os

from oslo_db.sqlalchemy import test_base
from oslo_db.sqlalchemy import utils as db_utils

from sahara.tests.unit.db.migration import test_migrations_base as base


class SaharaMigrationsCheckers(object):

    def assertColumnExists(self, engine, table, column):
        t = db_utils.get_table(engine, table)
        self.assertIn(column, t.c)

    def assertColumnsExist(self, engine, table, columns):
        for column in columns:
            self.assertColumnExists(engine, table, column)

    def assertColumnType(self, engine, table, column, column_type):
        t = db_utils.get_table(engine, table)
        column_ref_type = str(t.c[column].type)
        self.assertEqual(column_ref_type, column_type)

    def assertColumnCount(self, engine, table, columns):
        t = db_utils.get_table(engine, table)
        self.assertEqual(len(columns), len(t.columns))

    def assertColumnNotExists(self, engine, table, column):
        t = db_utils.get_table(engine, table)
        self.assertNotIn(column, t.c)

    def assertIndexExists(self, engine, table, index):
        t = db_utils.get_table(engine, table)
        index_names = [idx.name for idx in t.indexes]
        self.assertIn(index, index_names)

    def assertIndexMembers(self, engine, table, index, members):
        self.assertIndexExists(engine, table, index)

        t = db_utils.get_table(engine, table)
        index_columns = None
        for idx in t.indexes:
            if idx.name == index:
                index_columns = idx.columns.keys()
                break

        self.assertEqual(sorted(members), sorted(index_columns))

    def test_walk_versions(self):
        self.walk_versions(self.engine)

    def _pre_upgrade_001(self, engine):
        # Anything returned from this method will be
        # passed to corresponding _check_xxx method as 'data'.
        pass

    def _check_001(self, engine, data):
        job_binary_internal_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'name',
            'data',
            'datasize'
        ]
        self.assertColumnsExist(
            engine, 'job_binary_internal', job_binary_internal_columns)
        self.assertColumnCount(
            engine, 'job_binary_internal', job_binary_internal_columns)

        node_group_templates_columns = [
            'created_at',
            'updated_at',
            'id',
            'name',
            'description',
            'tenant_id',
            'flavor_id',
            'image_id',
            'plugin_name',
            'hadoop_version',
            'node_processes',
            'node_configs',
            'volumes_per_node',
            'volumes_size',
            'volume_mount_prefix',
            'floating_ip_pool'
        ]
        self.assertColumnsExist(
            engine, 'node_group_templates', node_group_templates_columns)
        self.assertColumnCount(
            engine, 'node_group_templates', node_group_templates_columns)

        data_sources_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'name',
            'description',
            'type',
            'url',
            'credentials'
        ]
        self.assertColumnsExist(
            engine, 'data_sources', data_sources_columns)
        self.assertColumnCount(
            engine, 'data_sources', data_sources_columns)

        cluster_templates_columns = [
            'created_at',
            'updated_at',
            'id',
            'name',
            'description',
            'cluster_configs',
            'default_image_id',
            'anti_affinity',
            'tenant_id',
            'neutron_management_network',
            'plugin_name',
            'hadoop_version'
        ]
        self.assertColumnsExist(
            engine, 'cluster_templates', cluster_templates_columns)
        self.assertColumnCount(
            engine, 'cluster_templates', cluster_templates_columns)

        job_binaries_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'name',
            'description',
            'url',
            'extra'
        ]
        self.assertColumnsExist(
            engine, 'job_binaries', job_binaries_columns)
        self.assertColumnCount(
            engine, 'job_binaries', job_binaries_columns)

        jobs_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'name',
            'description',
            'type'
        ]
        self.assertColumnsExist(engine, 'jobs', jobs_columns)
        self.assertColumnCount(engine, 'jobs', jobs_columns)

        templates_relations_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'name',
            'flavor_id',
            'image_id',
            'node_processes',
            'node_configs',
            'volumes_per_node',
            'volumes_size',
            'volume_mount_prefix',
            'count',
            'cluster_template_id',
            'node_group_template_id',
            'floating_ip_pool'
        ]
        self.assertColumnsExist(
            engine, 'templates_relations', templates_relations_columns)
        self.assertColumnCount(
            engine, 'templates_relations', templates_relations_columns)

        mains_association_columns = [
            'Job_id',
            'JobBinary_id'
        ]
        self.assertColumnsExist(
            engine, 'mains_association', mains_association_columns)
        self.assertColumnCount(
            engine, 'mains_association', mains_association_columns)

        libs_association_columns = [
            'Job_id',
            'JobBinary_id'
        ]
        self.assertColumnsExist(
            engine, 'libs_association', libs_association_columns)
        self.assertColumnCount(
            engine, 'libs_association', libs_association_columns)

        clusters_columns = [
            'created_at',
            'updated_at',
            'id',
            'name',
            'description',
            'tenant_id',
            'trust_id',
            'is_transient',
            'plugin_name',
            'hadoop_version',
            'cluster_configs',
            'default_image_id',
            'neutron_management_network',
            'anti_affinity',
            'management_private_key',
            'management_public_key',
            'user_keypair_id',
            'status',
            'status_description',
            'info',
            'extra',
            'cluster_template_id'
        ]
        self.assertColumnsExist(engine, 'clusters', clusters_columns)
        self.assertColumnCount(engine, 'clusters', clusters_columns)

        node_groups_columns = [
            'created_at',
            'updated_at',
            'id',
            'name',
            'tenant_id',
            'flavor_id',
            'image_id',
            'image_username',
            'node_processes',
            'node_configs',
            'volumes_per_node',
            'volumes_size',
            'volume_mount_prefix',
            'count',
            'cluster_id',
            'node_group_template_id',
            'floating_ip_pool'
        ]
        self.assertColumnsExist(engine, 'node_groups', node_groups_columns)
        self.assertColumnCount(engine, 'node_groups', node_groups_columns)

        job_executions_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'job_id',
            'input_id',
            'output_id',
            'start_time',
            'end_time',
            'cluster_id',
            'info',
            'progress',
            'oozie_job_id',
            'return_code',
            'job_configs',
            'extra'
        ]
        self.assertColumnsExist(
            engine, 'job_executions', job_executions_columns)
        self.assertColumnCount(
            engine, 'job_executions', job_executions_columns)

        instances_columns = [
            'created_at',
            'updated_at',
            'id',
            'tenant_id',
            'node_group_id',
            'instance_id',
            'instance_name',
            'internal_ip',
            'management_ip',
            'volumes'
        ]
        self.assertColumnsExist(engine, 'instances', instances_columns)
        self.assertColumnCount(engine, 'instances', instances_columns)

        self._data_001(engine, data)

    def _data_001(self, engine, data):
        datasize = 512 * 1024  # 512kB
        data = os.urandom(datasize)
        t = db_utils.get_table(engine, 'job_binary_internal')
        engine.execute(t.insert(), data=data, id='123', name='name')
        new_data = engine.execute(t.select()).fetchone().data
        self.assertEqual(data, new_data)
        engine.execute(t.delete())

    def _check_002(self, engine, data):
        # currently, 002 is just a placeholder
        pass

    def _check_003(self, engine, data):
        # currently, 003 is just a placeholder
        pass

    def _check_004(self, engine, data):
        # currently, 004 is just a placeholder
        pass

    def _check_005(self, engine, data):
        # currently, 005 is just a placeholder
        pass

    def _check_006(self, engine, data):
        # currently, 006 is just a placeholder
        pass

    def _pre_upgrade_007(self, engine):
        desc = 'magic'
        t = db_utils.get_table(engine, 'clusters')
        engine.execute(t.insert(), id='123', name='name', plugin_name='pname',
                       hadoop_version='1', management_private_key='2',
                       management_public_key='3', status_description=desc)

    def _check_007(self, engine, data):
        t = db_utils.get_table(engine, 'clusters')
        res = engine.execute(t.select().where(t.c.id == '123')).first()
        self.assertEqual('magic', res['status_description'])
        engine.execute(t.delete())

        # check that status_description can keep 128kb.
        # MySQL varchar can not keep more then 64kb
        desc = 'a' * 128 * 1024  # 128kb
        t = db_utils.get_table(engine, 'clusters')
        engine.execute(t.insert(), id='123', name='name', plugin_name='plname',
                       hadoop_version='hversion', management_private_key='1',
                       management_public_key='2', status_description=desc)
        new_desc = engine.execute(t.select()).fetchone().status_description
        self.assertEqual(desc, new_desc)
        engine.execute(t.delete())

    def _check_008(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates',
                                'security_groups')
        self.assertColumnExists(engine, 'node_groups', 'security_groups')
        self.assertColumnExists(engine, 'templates_relations',
                                'security_groups')

    def _check_009(self, engine, data):
        self.assertColumnExists(engine, 'clusters', 'rollback_info')

    def _check_010(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates',
                                'auto_security_group')
        self.assertColumnExists(engine, 'node_groups', 'auto_security_group')
        self.assertColumnExists(engine, 'templates_relations',
                                'auto_security_group')
        self.assertColumnExists(engine, 'node_groups', 'open_ports')

    def _check_011(self, engine, data):
        self.assertColumnExists(engine, 'clusters', 'sahara_info')

    def _check_012(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates',
                                'availability_zone')
        self.assertColumnExists(engine, 'node_groups', 'availability_zone')
        self.assertColumnExists(engine, 'templates_relations',
                                'availability_zone')

    def _check_014(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates', 'volume_type')
        self.assertColumnExists(engine, 'node_groups', 'volume_type')
        self.assertColumnExists(engine, 'templates_relations', 'volume_type')

    def _check_015(self, engine, data):
        provision_steps_columns = [
            'created_at',
            'updated_at',
            'id',
            'cluster_id',
            'tenant_id',
            'step_name',
            'step_type',
            'completed',
            'total',
            'successful',
            'started_at',
            'completed_at',
        ]
        events_columns = [
            'created_at',
            'updated_at',
            'id',
            'node_group_id',
            'instance_id',
            'instance_name',
            'event_info',
            'successful',
            'step_id',
        ]

        self.assertColumnCount(engine, 'cluster_provision_steps',
                               provision_steps_columns)
        self.assertColumnsExist(engine, 'cluster_provision_steps',
                                provision_steps_columns)

        self.assertColumnCount(engine, 'cluster_events', events_columns)
        self.assertColumnsExist(engine, 'cluster_events', events_columns)

    def _check_016(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates',
                                'is_proxy_gateway')
        self.assertColumnExists(engine, 'node_groups', 'is_proxy_gateway')
        self.assertColumnExists(engine, 'templates_relations',
                                'is_proxy_gateway')

    def _check_017(self, engine, data):
        self.assertColumnNotExists(engine, 'job_executions', 'progress')

    def _check_018(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates',
                                'volume_local_to_instance')
        self.assertColumnExists(engine, 'node_groups',
                                'volume_local_to_instance')
        self.assertColumnExists(engine, 'templates_relations',
                                'volume_local_to_instance')

    def _check_019(self, engine, data):
        self.assertColumnExists(engine, 'node_group_templates', 'is_default')
        self.assertColumnExists(engine, 'cluster_templates', 'is_default')

    def _check_020(self, engine, data):
        self.assertColumnNotExists(engine, 'cluster_provision_steps',
                                   'completed')
        self.assertColumnNotExists(engine, 'cluster_provision_steps',
                                   'completed_at')
        self.assertColumnNotExists(engine, 'cluster_provision_steps',
                                   'started_at')

    def _check_021(self, engine, data):
        self.assertColumnExists(engine, 'job_executions', 'data_source_urls')

    def _check_022(self, engine, data):
        columns = [
            'created_at',
            'updated_at',
            'id',
            'job_id',
            'tenant_id',
            'name',
            'description',
            'mapping_type',
            'location',
            'value_type',
            'required',
            'order',
            'default'
        ]

        self.assertColumnCount(engine, 'job_interface_arguments', columns)
        self.assertColumnsExist(engine, 'job_interface_arguments', columns)

    def _check_023(self, engine, data):
        self.assertColumnExists(engine, 'clusters',
                                'use_autoconfig')
        self.assertColumnExists(engine, 'cluster_templates',
                                'use_autoconfig')
        self.assertColumnExists(engine, 'node_group_templates',
                                'use_autoconfig')
        self.assertColumnExists(engine, 'node_groups',
                                'use_autoconfig')
        self.assertColumnExists(engine, 'templates_relations',
                                'use_autoconfig')

    def _check_024(self, engine, data):
        tables = [
            'node_group_templates',
            'node_groups',
            'templates_relations',
            'clusters',
            'cluster_templates'
        ]

        for table in tables:
            self.assertColumnExists(engine, table, 'shares')

    def _check_025(self, engine, data):
        self.assertColumnType(engine, 'instances', 'internal_ip',
                              'VARCHAR(45)')
        self.assertColumnType(engine, 'instances', 'management_ip',
                              'VARCHAR(45)')

    def _check_026(self, engine, data):
        tables = [
            'clusters',
            'cluster_templates',
            'node_group_templates',
            'data_sources',
            'job_executions',
            'jobs',
            'job_binary_internal',
            'job_binaries',
        ]
        for table in tables:
            self.assertColumnExists(engine, table, 'is_public')
            self.assertColumnExists(engine, table, 'is_protected')

    def _check_027(self, engine, data):
        self.assertColumnNotExists(engine, 'job_executions',
                                   'oozie_job_id')
        self.assertColumnExists(engine, 'job_executions',
                                'engine_job_id')

    def _check_028(self, engine, data):
        self.assertColumnExists(engine, 'instances', 'storage_devices_number')

    def _pre_upgrade_029(self, engine):
        t = db_utils.get_table(engine, 'node_group_templates')
        engine.execute(t.insert(), id='123', name='first', plugin_name='plg',
                       hadoop_version='1', flavor_id='1', volumes_per_node=0,
                       is_default=True, is_protected=False)

        engine.execute(t.insert(), id='124', name='second', plugin_name='plg',
                       hadoop_version='1', flavor_id='1', volumes_per_node=0,
                       is_default=False, is_protected=False)

        t = db_utils.get_table(engine, 'cluster_templates')
        engine.execute(t.insert(), id='123', name='name', plugin_name='plg',
                       hadoop_version='1', is_default=True, is_protected=False)

        engine.execute(t.insert(), id='124', name='name', plugin_name='plg',
                       hadoop_version='1', is_default=False,
                       is_protected=False)

    def _check_029(self, engine, data):
        t = db_utils.get_table(engine, 'node_group_templates')
        res = engine.execute(t.select().where(t.c.id == '123')).first()
        self.assertTrue(res['is_protected'])

        res = engine.execute(t.select().where(t.c.id == '124')).first()
        self.assertFalse(res['is_protected'])
        engine.execute(t.delete())

        t = db_utils.get_table(engine, 'cluster_templates')
        res = engine.execute(t.select().where(t.c.id == '123')).first()
        self.assertTrue(res['is_protected'])

        res = engine.execute(t.select().where(t.c.id == '124')).first()
        self.assertFalse(res['is_protected'])
        engine.execute(t.delete())

    def _check_030(self, engine, data):
        health_check_columns = [
            'status',
            'name',
            'description',
            'id',
            'verification_id',
            'created_at',
            'updated_at'
        ]

        verification_columns = [
            'status',
            'id',
            'cluster_id',
            'created_at',
            'updated_at'
        ]

        self.assertColumnCount(engine, 'cluster_verifications',
                               verification_columns)
        self.assertColumnsExist(engine, 'cluster_verifications',
                                verification_columns)

        self.assertColumnCount(engine, 'cluster_health_checks',
                               health_check_columns)
        self.assertColumnsExist(engine, 'cluster_health_checks',
                                health_check_columns)

    def _check_031(self, engine, data):
        plugins_data_columns = [
            'name',
            'id',
            'tenant_id',
            'version_labels',
            'plugin_labels',
            'updated_at',
            'created_at'
        ]

        self.assertColumnCount(engine, 'plugin_data',
                               plugins_data_columns)
        self.assertColumnsExist(engine, 'plugin_data',
                                plugins_data_columns)

    def _check_033(self, engine, data):
        self.assertColumnExists(engine, 'clusters', 'anti_affinity_ratio')

    def _check_034(self, engine, data):
        self.assertColumnExists(engine, 'node_groups',
                                'boot_from_volume')
        self.assertColumnExists(engine, 'node_group_templates',
                                'boot_from_volume')
        self.assertColumnExists(engine, 'templates_relations',
                                'boot_from_volume')

    def _check_035(self, engine, data):
        for col in ['boot_volume_type',
                    'boot_volume_availability_zone',
                    'boot_volume_local_to_instance']:
            self.assertColumnExists(engine, 'node_groups', col)
            self.assertColumnExists(engine, 'node_group_templates', col)
            self.assertColumnExists(engine, 'templates_relations', col)


class TestMigrationsMySQL(SaharaMigrationsCheckers,
                          base.BaseWalkMigrationTestCase,
                          base.TestModelsMigrationsSync,
                          test_base.MySQLOpportunisticTestCase):
    pass


class TestMigrationsPostgresql(SaharaMigrationsCheckers,
                               base.BaseWalkMigrationTestCase,
                               base.TestModelsMigrationsSync,
                               test_base.PostgreSQLOpportunisticTestCase):
    pass



# Copyright (c) 2014 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from sahara.utils import edp


class EdpUtilTest(testtools.TestCase):
    def test_split_job_type(self):
        jtype, stype = edp.split_job_type(edp.JOB_TYPE_MAPREDUCE)
        self.assertEqual(edp.JOB_TYPE_MAPREDUCE, jtype)
        self.assertEqual(edp.JOB_SUBTYPE_NONE, stype)

        jtype, stype = edp.split_job_type(edp.JOB_TYPE_MAPREDUCE_STREAMING)
        self.assertEqual(edp.JOB_TYPE_MAPREDUCE, jtype)
        self.assertEqual(edp.JOB_SUBTYPE_STREAMING, stype)

    def test_compare_job_type(self):
        self.assertTrue(edp.compare_job_type(
            edp.JOB_TYPE_JAVA,
            edp.JOB_TYPE_JAVA,
            edp.JOB_TYPE_MAPREDUCE,
            strict=True))
        self.assertFalse(edp.compare_job_type(
            edp.JOB_TYPE_MAPREDUCE_STREAMING,
            edp.JOB_TYPE_JAVA,
            edp.JOB_TYPE_MAPREDUCE,
            strict=True))
        self.assertTrue(edp.compare_job_type(
            edp.JOB_TYPE_MAPREDUCE_STREAMING,
            edp.JOB_TYPE_JAVA,
            edp.JOB_TYPE_MAPREDUCE))
        self.assertFalse(edp.compare_job_type(
            edp.JOB_TYPE_MAPREDUCE,
            edp.JOB_TYPE_JAVA,
            edp.JOB_TYPE_MAPREDUCE_STREAMING))

    def test_get_builtin_binaries_java_available(self):
        job = mock.Mock(type=edp.JOB_TYPE_JAVA)
        configs = {edp.ADAPT_FOR_OOZIE: True}
        binaries = edp.get_builtin_binaries(job, configs)
        self.assertEqual(1, len(binaries))
        binary = binaries[0]
        self.assertTrue(binary['name'].startswith('builtin-'))
        self.assertTrue(binary['name'].endswith('.jar'))
        self.assertIsNotNone(binary['raw'])

    def test_get_builtin_binaries_empty(self):
        for job_type in edp.JOB_TYPES_ALL:
            job = mock.Mock(type=job_type)
            self.assertEqual(0, len(edp.get_builtin_binaries(job, {})))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from sahara import exceptions as ex
from sahara.i18n import _
from sahara.plugins import base as plugins_base
from sahara.utils import resources


class ProvisioningPluginBase(plugins_base.PluginInterface):
    @plugins_base.required
    def get_versions(self):
        """Get available plugin versions

        :returns: A sequence of strings representing the versions

        For example:
            ["1.0.0", "1.0.1"]
        """
        pass

    @plugins_base.required
    def get_configs(self, hadoop_version):
        """Get default configuration for a given plugin version

        :param hadoop_version: String representing a plugin version
        :returns: A dict containing the configuration
        """
        pass

    @plugins_base.required_with_default
    def get_labels(self):
        versions = self.get_versions()
        default = {'enabled': {'status': True}}
        return {
            'plugin_labels': copy.deepcopy(default),
            'version_labels': {
                version: copy.deepcopy(default) for version in versions
            }
        }

    @plugins_base.required
    def get_node_processes(self, hadoop_version):
        """Get node processes of a given plugin version

        :param hadoop_version: String containing a plugin version
        :returns: A dict where the keys are the core components of the plugin
        and the value is a sequence of node processes for that component

        For example:
        {
            "HDFS": ["namenode", "datanode"],
            "Spark": ["master", "slave"]
        }

        """
        pass

    @plugins_base.required_with_default
    def get_required_image_tags(self, hadoop_version):
        return [self.name, hadoop_version]

    @plugins_base.required_with_default
    def validate(self, cluster):
        pass

    @plugins_base.required_with_default
    def validate_scaling(self, cluster, existing, additional):
        pass

    @plugins_base.required_with_default
    def update_infra(self, cluster):
        pass

    @plugins_base.required
    def configure_cluster(self, cluster):
        pass

    @plugins_base.required
    def start_cluster(self, cluster):
        pass

    @plugins_base.optional
    def scale_cluster(self, cluster, instances):
        pass

    @plugins_base.optional
    def get_edp_engine(self, cluster, job_type):
        pass

    @plugins_base.optional
    def get_edp_job_types(self, versions=None):
        return {}

    @plugins_base.optional
    def get_edp_config_hints(self, job_type, version):
        return {}

    @plugins_base.required_with_default
    def get_open_ports(self, node_group):
        return []

    @plugins_base.required_with_default
    def decommission_nodes(self, cluster, instances):
        pass

    @plugins_base.optional
    def get_image_arguments(self, hadoop_version):
        """Gets the argument set taken by the plugin's image generator

        Note: If the plugin can generate or validate an image but takes no
        arguments, please return an empty sequence rather than NotImplemented
        for all versions that support image generation or validation. This is
        used as a flag to determine whether the plugin has implemented this
        optional feature.

        :returns: A sequence with items of type
            sahara.plugins.images.ImageArgument
        """
        return NotImplemented

    @plugins_base.optional
    def pack_image(self, hadoop_version, remote,
                   test_only=False, image_arguments=None):
        """Packs an image for registration in Glance and use by Sahara

        :param remote: A remote (usually of type
            sahara.cli.image_pack.api.ImageRemote) that serves as a handle to
            the image to modify. Note that this image will be modified
            in-place, not copied.
        :param test_only: If set to True, this method will only test to
            ensure that the image already meets the plugin's requirements.
            This can be used to test images without modification. If set to
            False per the default, this method will modify the image if any
            requirements are not met.
        :param image_arguments: A dict of image argument name to argument
            value.
        :raises: sahara.plugins.exceptions.ImageValidationError: If the method
            fails to modify the image to specification (if test_only is False),
            or if the method finds that the image does not meet the
            specification (if test_only is True).
        :raises: sahara.plugins.exceptions.ImageValidationSpecificationError:
            If the specification for image generation or validation is itself
            in error and cannot be executed without repair.
        """
        pass

    @plugins_base.optional
    def validate_images(self, cluster, test_only=False, image_arguments=None):
        """Validates the image to be used by a cluster.

        :param cluster: The object handle to a cluster which has active
            instances ready to generate remote handles.
        :param test_only: If set to True, this method will only test to
            ensure that the image already meets the plugin's requirements.
            This can be used to test images without modification. If set to
            False per the default, this method will modify the image if any
            requirements are not met.
        :param image_arguments: A dict of image argument name to argument
            value.
        :raises: sahara.plugins.exceptions.ImageValidationError: If the method
            fails to modify the image to specification (if test_only is False),
            or if the method finds that the image does not meet the
            specification (if test_only is True).
        :raises: sahara.plugins.exceptions.ImageValidationSpecificationError:
            If the specification for image generation or validation is itself
            in error and cannot be executed without repair.
        """
        pass

    @plugins_base.required_with_default
    def on_terminate_cluster(self, cluster):
        pass

    @plugins_base.optional
    def recommend_configs(self, cluster, scaling=False):
        pass

    @plugins_base.required_with_default
    def get_health_checks(self, cluster):
        return []

    def get_all_configs(self, hadoop_version):
        common = list_of_common_configs()
        plugin_specific_configs = self.get_configs(hadoop_version)
        if plugin_specific_configs:
            common.extend(plugin_specific_configs)
        return common

    def get_version_details(self, version):
        details = {}
        configs = self.get_all_configs(version)
        details['configs'] = [c.dict for c in configs]
        details['node_processes'] = self.get_node_processes(version)
        details['required_image_tags'] = self.get_required_image_tags(version)
        return details

    def to_dict(self):
        res = super(ProvisioningPluginBase, self).to_dict()
        res['versions'] = self.get_versions()
        return res

    # Some helpers for plugins

    def _map_to_user_inputs(self, hadoop_version, configs):
        config_objs = self.get_all_configs(hadoop_version)

        # convert config objects to applicable_target -> config_name -> obj
        config_objs_map = {}
        for config_obj in config_objs:
            applicable_target = config_obj.applicable_target
            confs = config_objs_map.get(applicable_target, {})
            confs[config_obj.name] = config_obj
            config_objs_map[applicable_target] = confs

        # iterate over all configs and append UserInputs to result list
        result = []
        for applicable_target in configs:
            for config_name in configs[applicable_target]:
                confs = config_objs_map.get(applicable_target)
                if not confs:
                    raise ex.ConfigurationError(
                        _("Can't find applicable target "
                          "'%(applicable_target)s' for '%(config_name)s'")
                        % {"applicable_target": applicable_target,
                           "config_name": config_name})
                conf = confs.get(config_name)
                if not conf:
                    raise ex.ConfigurationError(
                        _("Can't find config '%(config_name)s' "
                          "in '%(applicable_target)s'")
                        % {"config_name": config_name,
                           "applicable_target": applicable_target})
                result.append(UserInput(
                    conf, configs[applicable_target][config_name]))

        return sorted(result)


class Config(resources.BaseResource):
    """Describes a single config parameter.

    Config type - could be 'str', 'integer', 'boolean', 'enum'.
    If config type is 'enum' then list of valid values should be specified in
    config_values property.

    Priority - integer parameter which helps to differentiate all
    configurations in the UI. Priority decreases from the lower values to
    higher values.

    For example:

        "some_conf", "map_reduce", "node", is_optional=True
    """

    def __init__(self, name, applicable_target, scope, config_type="string",
                 config_values=None, default_value=None, is_optional=False,
                 description=None, priority=2):
        self.name = name
        self.description = description
        self.config_type = config_type
        self.config_values = config_values
        self.default_value = default_value
        self.applicable_target = applicable_target
        self.scope = scope
        self.is_optional = is_optional
        self.priority = priority

    def to_dict(self):
        res = super(Config, self).to_dict()
        # TODO(slukjanov): all custom fields from res
        return res

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return '<Config %s in %s>' % (self.name, self.applicable_target)


class UserInput(object):
    """Value provided by the user for a specific config entry."""

    def __init__(self, config, value):
        self.config = config
        self.value = value

    def __eq__(self, other):
        return self.config == other.config and self.value == other.value

    def __lt__(self, other):
        return (self.config, self.value) < (other.config, other.value)

    def __repr__(self):
        return '<UserInput %s = %s>' % (self.config.name, self.value)


class ValidationError(object):
    """Describes what is wrong with one of the values provided by user."""

    def __init__(self, config, message):
        self.config = config
        self.message = message

    def __repr__(self):
        return "<ValidationError %s>" % self.config.name


# COMMON FOR ALL PLUGINS CONFIGS

XFS_ENABLED = Config(
    "Enable XFS", 'general', 'cluster', priority=1,
    default_value=True, config_type="bool", is_optional=True,
    description='Enables XFS for formatting'
)

DISKS_PREPARING_TIMEOUT = Config(
    "Timeout for disk preparing", 'general', 'cluster', priority=1,
    default_value=300, config_type="int", is_optional=True,
    description='Timeout for preparing disks, formatting and mounting'
)


NTP_URL = Config(
    "URL of NTP server", 'general', 'cluster', priority=1,
    default_value='', is_optional=True,
    description='URL of the NTP server for synchronization time on cluster'
                ' instances'
)

NTP_ENABLED = Config(
    "Enable NTP service", 'general', 'cluster', priority=1, default_value=True,
    config_type="bool",
    description='Enables NTP service for synchronization time on cluster '
                'instances'
)

HEAT_WAIT_CONDITION_TIMEOUT = Config(
    "Heat Wait Condition timeout", "general", "cluster", priority=1,
    config_type="int", default_value=3600, is_optional=True,
    description="The number of seconds to wait for the instance to boot")


def list_of_common_configs():
    return [DISKS_PREPARING_TIMEOUT, NTP_ENABLED, NTP_URL,
            HEAT_WAIT_CONDITION_TIMEOUT, XFS_ENABLED]



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from sahara.api import acl
from sahara.service.api.v2 import jobs as api
from sahara.service import validation as v
from sahara.service.validations.edp import job_execution as v_j_e
from sahara.service.validations.edp import job_execution_schema as v_j_e_schema
import sahara.utils.api as u


rest = u.RestV2('jobs', __name__)


def _replace_job_id_job_template_id(job_obj):
    dict.update(job_obj, {'job_template_id': job_obj['job_id']})
    dict.pop(job_obj, 'job_id')


@rest.get('/jobs')
@acl.enforce("data-processing:job:list")
@v.check_exists(api.get_job_execution, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_job_executions)
@v.validate_request_params(['status'])
def jobs_list():
    result = api.job_execution_list(**u.get_request_args().to_dict())
    # APIv2: renaming oozie_job_id -> engine_job_id
    # once APIv1 is deprecated this can be
    # removed
    for je in result:
        je.pop('oozie_job_id', force=True)
        u._replace_tenant_id_project_id(je)
        _replace_job_id_job_template_id(je)
    return u.render(res=result, name='jobs')


@rest.post('/jobs')
@acl.enforce("data-processing:job:execute")
@v.validate(v_j_e_schema.JOB_EXEC_SCHEMA_V2, v_j_e.check_job_execution)
@v.validate_request_params([])
def jobs_execute(data):
    result = {'job': api.execute_job(data)}
    dict.update(result['job'],
                {'engine_job_id': result['job']['oozie_job_id']})
    dict.pop(result['job'], 'oozie_job_id')
    u._replace_tenant_id_project_id(result['job'])
    _replace_job_id_job_template_id(result['job'])
    return u.render(result)


@rest.get('/jobs/<job_id>')
@acl.enforce("data-processing:job:get")
@v.check_exists(api.get_job_execution, id='job_id')
@v.validate_request_params([])
def jobs_get(job_id):
    data = u.get_request_args()
    refresh_status = six.text_type(
        data.get('refresh_status', 'false')).lower() == 'true'
    result = {'job': api.get_job_execution(job_id, refresh_status)}
    result['job'].pop('oozie_job_id', force=True)
    u._replace_tenant_id_project_id(result['job'])
    _replace_job_id_job_template_id(result['job'])
    return u.render(result)


@rest.patch('/jobs/<job_id>')
@acl.enforce("data-processing:job:update")
@v.check_exists(api.get_job_execution, id='job_id')
@v.validate(
    v_j_e_schema.JOB_EXEC_UPDATE_SCHEMA, v_j_e.check_job_execution_update)
@v.validate_request_params([])
def jobs_update(job_id, data):
    result = {'job': api.update_job_execution(job_id, data)}
    result['job'].pop('oozie_job_id', force=True)
    u._replace_tenant_id_project_id(result['job'])
    _replace_job_id_job_template_id(result['job'])
    return u.render(result)


@rest.delete('/jobs/<job_id>')
@acl.enforce("data-processing:job:delete")
@v.check_exists(api.get_job_execution, id='job_id')
@v.validate(None, v_j_e.check_job_execution_delete)
@v.validate_request_params([])
def jobs_delete(job_id):
    api.delete_job_execution(job_id)
    return u.render()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from sahara import exceptions as ex
from sahara.utils.openstack import heat as h


def stack(status, upd_time=None):
    status_reason = status
    status = status[status.index('_') + 1:]
    return mock.Mock(status=status, updated_time=upd_time,
                     stack_status_reason=status_reason)


class TestClusterStack(testtools.TestCase):
    @mock.patch('sahara.utils.openstack.heat.get_stack')
    @mock.patch("sahara.context.sleep", return_value=None)
    def test_wait_completion(self, sleep, client):
        cl = mock.Mock(stack_name='cluster')
        client.side_effect = [stack(
            'CREATE_IN_PROGRESS'), stack('CREATE_COMPLETE')]
        h.wait_stack_completion(cl)
        self.assertEqual(2, client.call_count)

        client.side_effect = [
            stack('UPDATE_IN_PROGRESS'), stack('UPDATE_COMPLETE')]
        h.wait_stack_completion(cl)
        self.assertEqual(4, client.call_count)

        client.side_effect = [
            stack('DELETE_IN_PROGRESS'), stack('DELETE_COMPLETE')]
        h.wait_stack_completion(cl)
        self.assertEqual(6, client.call_count)

        client.side_effect = [
            stack('CREATE_COMPLETE'), stack('CREATE_COMPLETE'),
            stack('UPDATE_IN_PROGRESS'), stack('UPDATE_COMPLETE', 1)]

        h.wait_stack_completion(cl, is_update=True)
        self.assertEqual(10, client.call_count)

        client.side_effect = [stack('UPDATE_COMPLETE'), stack(
            'UPDATE_IN_PROGRESS'), stack('UPDATE_COMPLETE', 1)]
        h.wait_stack_completion(cl, is_update=True)
        self.assertEqual(13, client.call_count)

        client.side_effect = [
            stack('CREATE_IN_PROGRESS'), stack('CREATE_FAILED')]
        with testtools.ExpectedException(
                ex.HeatStackException,
                value_re=("Heat stack failed with status "
                          "CREATE_FAILED\nError ID: .*")):
            h.wait_stack_completion(cl)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from sahara.utils import types


class TypesTestCase(testtools.TestCase):

    def test_is_int(self):
        self.assertTrue(types.is_int('1'))
        self.assertTrue(types.is_int('0'))
        self.assertTrue(types.is_int('-1'))
        self.assertFalse(types.is_int('1.1'))
        self.assertFalse(types.is_int('ab'))
        self.assertFalse(types.is_int(''))



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara import conductor  # noqa
from sahara.plugins import base as plugins_base
from sahara.utils import remote

try:
    import guestfs
except ImportError:
    raise Exception("The image packing API depends on the system package "
                    "python-libguestfs (and libguestfs itself.) Please "
                    "install these packages to proceed.")


LOG = None
CONF = None


# This is broken out to support testability
def set_logger(log):
    global LOG
    LOG = log


# This is broken out to support testability
def set_conf(conf):
    global CONF
    CONF = conf


# This is a local exception class that is used to exit routines
# in cases where error information has already been logged.
# It is caught and suppressed everywhere it is used.
class Handled(Exception):
    pass


class Context(object):
    '''Create a pseudo Context object

    Since this tool does not use the REST interface, we
    do not have a request from which to build a Context.
    '''
    def __init__(self, is_admin=False, tenant_id=None):
        self.is_admin = is_admin
        self.tenant_id = tenant_id


class ImageRemote(remote.TerminalOnlyRemote):

    def __init__(self, image_path, root_drive):
        guest = guestfs.GuestFS(python_return_dict=True)
        guest.add_drive_opts(image_path, format="qcow2")
        guest.set_network(True)
        self.guest = guest
        self.root_drive = root_drive

    def __enter__(self):
        self.guest.launch()
        if not self.root_drive:
            self.root_drive = self.guest.inspect_os()[0]
        self.guest.mount(self.root_drive, '/')
        try:
            cmd = "echo Testing sudo without tty..."
            self.execute_command(cmd, run_as_root=True)
        except RuntimeError:
            cmd = "sed -i 's/requiretty/!requiretty/' /etc/sudoers"
            self.guest.execute_command(cmd)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.guest.sync()
        self.guest.umount_all()
        self.guest.close()

    def execute_command(self, cmd, run_as_root=False, get_stderr=False,
                        raise_when_error=True, timeout=300):
        try:
            LOG.info("Issuing command: {cmd}".format(cmd=cmd))
            stdout = self.guest.sh(cmd)
            LOG.info("Received response: {stdout}".format(stdout=stdout))
            return 0, stdout
        except RuntimeError as ex:
            if raise_when_error:
                raise
            else:
                return 1, ex.message

    def get_os_distrib(self):
        return self.guest.inspect_get_distro(self.root_drive)

    def write_file_to(self, path, script, run_as_root):
        LOG.info("Writing script to : {path}".format(path=path))
        stdout = self.guest.write(path, script)
        return 0, stdout


def setup_plugins():
    plugins_base.setup_plugins()


def get_loaded_plugins():
    return plugins_base.PLUGINS.plugins


def get_plugin_arguments(plugin_name):
    """Gets plugin arguments, as a dict of version to argument list."""
    plugin = plugins_base.PLUGINS.get_plugin(plugin_name)
    versions = plugin.get_versions()
    return {version: plugin.get_image_arguments(version)
            for version in versions}


def pack_image(image_path, plugin_name, plugin_version, image_arguments,
               root_drive=None, test_only=False):
    with ImageRemote(image_path, root_drive) as image_remote:
        plugin = plugins_base.PLUGINS.get_plugin(plugin_name)
        plugin.pack_image(plugin_version, image_remote, test_only=test_only,
                          image_arguments=image_arguments)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import testtools

from sahara import exceptions as ex
from sahara.service.api import v10 as api
from sahara.service.health import verification_base
from sahara.service.validations import clusters as c_val
from sahara.service.validations import clusters_schema as c_schema
from sahara.tests.unit.service.validation import utils as u
from sahara.tests.unit import testutils as tu


class TestClusterUpdateValidation(u.ValidationTestCase):
    def setUp(self):
        super(TestClusterUpdateValidation, self).setUp()
        self._create_object_fun = mock.Mock()
        self.scheme = c_schema.CLUSTER_UPDATE_SCHEMA
        api.plugin_base.setup_plugins()

    def test_cluster_update_types(self):
        self._assert_types({
            'name': 'cluster',
            'description': 'very big cluster',
            'is_public': False,
            'is_protected': False,
            'shares': []
        })

    def test_cluster_update_nothing_required(self):
        self._assert_create_object_validation(
            data={}
        )

    def test_cluster_update(self):
        self._assert_create_object_validation(
            data={
                'name': 'cluster',
                'description': 'very big cluster',
                'is_public': False,
                'is_protected': False,
                'shares': []
            }
        )

        self._assert_create_object_validation(
            data={
                'name': 'cluster',
                'id': '1'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "Additional properties are not allowed "
                       "('id' was unexpected)")
        )

    @mock.patch('sahara.service.api.v10.get_cluster')
    def test_cluster_update_when_protected(self, get_cluster_p):
        cluster = tu.create_cluster("cluster1", "tenant_1", "fake",
                                    "0.1", ['ng1'], is_protected=True)
        get_cluster_p.return_value = cluster

        # cluster can't be updated if it's marked as protected
        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                c_val.check_cluster_update(cluster.id, {'name': 'new'})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e

        # cluster can be updated because is_protected flag was set to False
        c_val.check_cluster_update(
            cluster.id, {'is_protected': False, 'name': 'new'})

    @mock.patch('sahara.service.api.v10.get_cluster')
    def test_public_cluster_update_from_another_tenant(self, get_cluster_p):
        cluster = tu.create_cluster("cluster1", "tenant_2", "fake",
                                    "0.1", ['ng1'], is_public=True)
        get_cluster_p.return_value = cluster

        # cluster can't be updated from another tenant
        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                c_val.check_cluster_update(cluster.id, {'name': 'new'})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

    @mock.patch('sahara.conductor.API.cluster_get')
    def test_verifications_ops(self, get_cluster_mock):
        cluster = tu.create_cluster(
            'cluster1', "tenant_1", "fake", "0.1", ['ng1'], status='Active')
        get_cluster_mock.return_value = cluster
        self.assertIsNone(c_val.check_cluster_update(
            cluster, {'verification': {'status': "START"}}))
        cluster = tu.create_cluster(
            'cluster1', "tenant_1", "fake", "0.1", ['ng1'],
            status='Active', verification={'status': "CHECKING"})
        get_cluster_mock.return_value = cluster
        with testtools.ExpectedException(verification_base.CannotVerifyError):
            c_val.check_cluster_update(
                cluster, {'verification': {'status': 'START'}})

        cluster = tu.create_cluster(
            'cluster1', "tenant_1", "fake", "0.1", ['ng1'],
            status='Active', verification={'status': "RED"})
        get_cluster_mock.return_value = cluster
        self.assertIsNone(c_val.check_cluster_update(
            cluster, {'verification': {'status': "START"}}))

        with testtools.ExpectedException(verification_base.CannotVerifyError):
            c_val.check_cluster_update(cluster, {
                'is_public': True, 'verification': {'status': "START"}})

        # allow verification for protected resource
        cluster = tu.create_cluster(
            'cluster1', "tenant_1", "fake", "0.1", ['ng1'],
            is_protected=True, status='Active')
        get_cluster_mock.return_value = cluster
        self.assertIsNone(c_val.check_cluster_update(
            cluster, {'verification': {'status': "START"}}))
        # just for sure that protected works nicely for other
        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                c_val.check_cluster_update(cluster.id, {'name': 'new'})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os

import jsonschema
from oslo_config import cfg
from oslo_serialization import jsonutils as json
from oslo_utils import uuidutils
import six

from sahara import conductor
from sahara.db.templates import utils as u
from sahara.service.validations import cluster_template_schema as clt
from sahara.service.validations import node_group_template_schema as ngt
from sahara.utils import api_validator


LOG = None
CONF = None


# This is broken out to support testability
def set_logger(log):
    global LOG
    LOG = log


# This is broken out to support testability
def set_conf(conf):
    global CONF
    CONF = conf


ng_validator = api_validator.ApiValidator(ngt.NODE_GROUP_TEMPLATE_SCHEMA)
ct_validator = api_validator.ApiValidator(clt.CLUSTER_TEMPLATE_SCHEMA)

# Options that we allow to be replaced in a node group template
node_group_template_opts = [
    cfg.StrOpt('image_id',
               help='Image id field for a node group template.'),

    cfg.StrOpt('flavor_id',
               help='Flavor id field for a node group template.'),

    cfg.StrOpt('floating_ip_pool',
               help='Floating ip pool field for a node group template.'),
    cfg.BoolOpt('auto_security_group',
                default=False,
                help='Auto security group field for node group template.'),
    cfg.ListOpt('security_groups',
                default=[],
                help='Security group field for node group template.')
]

# Options that we allow to be replaced in a cluster template
cluster_template_opts = [
    cfg.StrOpt('default_image_id',
               help='Default image id field for a cluster template.'),

    cfg.StrOpt('neutron_management_network',
               help='Neutron management network '
               'field for a cluster template.')]

all_template_opts = node_group_template_opts + cluster_template_opts

node_group_template_opt_names = [o.name for o in node_group_template_opts]
cluster_template_opt_names = [o.name for o in cluster_template_opts]


# This is a local exception class that is used to exit routines
# in cases where error information has already been logged.
# It is caught and suppressed everywhere it is used.
class Handled(Exception):
    pass


class Context(object):
    '''Create a pseudo Context object

    Since this tool does not use the REST interface, we
    do not have a request from which to build a Context.
    '''
    def __init__(self, is_admin=False, tenant_id=None):
        self.is_admin = is_admin
        self.tenant_id = tenant_id


def check_usage_of_existing(ctx, ng_templates, cl_templates):
    '''Determine if any of the specified templates are in use

    This method searches for the specified templates by name and
    determines whether or not any existing templates are in use
    by a cluster or cluster template. Returns True if any of
    the templates are in use.

    :param ng_templates: A list of dictionaries. Each dictionary
                         has a "template" entry that represents
                         a node group template.
    :param cl_templates: A list of dictionaries. Each dictionary
                         has a "template" entry that represents
                         a cluster template
    :returns: True if any of the templates are in use, False otherwise
    '''
    error = False
    clusters = conductor.API.cluster_get_all(ctx)

    for ng_info in ng_templates:
        ng = u.find_node_group_template_by_name(ctx,
                                                ng_info["template"]["name"])
        if ng:
            cluster_users, template_users = u.check_node_group_template_usage(
                ng["id"], clusters)

            if cluster_users:
                LOG.warning("Node group template {name} "
                            "in use by clusters {clusters}".format(
                                name=ng["name"], clusters=cluster_users))
            if template_users:
                LOG.warning("Node group template {name} "
                            "in use by cluster templates {templates}".format(
                                name=ng["name"], templates=template_users))

            if cluster_users or template_users:
                LOG.warning("Update of node group template "
                            "{name} is not allowed".format(name=ng["name"]))
                error = True

    for cl_info in cl_templates:
        cl = u.find_cluster_template_by_name(ctx, cl_info["template"]["name"])
        if cl:
            cluster_users = u.check_cluster_template_usage(cl["id"], clusters)

            if cluster_users:
                LOG.warning("Cluster template {name} "
                            "in use by clusters {clusters}".format(
                                name=cl["name"], clusters=cluster_users))

                LOG.warning("Update of cluster template "
                            "{name} is not allowed".format(name=cl["name"]))
                error = True

    return error


def log_skipping_dir(path, reason=""):
    if reason:
        reason = ", " + reason
    LOG.warning("Skipping processing for {dir}{reason}".format(
        dir=path, reason=reason))


def check_cluster_templates_valid(ng_templates, cl_templates):
    # Check that if name references to node group templates
    # are replaced with a uuid value that the cluster template
    # passes JSON validation. We don't have the real uuid yet,
    # but this will allow the validation test.
    if ng_templates:
        dummy_uuid = uuidutils.generate_uuid()
        ng_ids = {ng["template"]["name"]: dummy_uuid for ng in ng_templates}
    else:
        ng_ids = {}

    for cl in cl_templates:
        template = copy.deepcopy(cl["template"])
        u.substitute_ng_ids(template, ng_ids)
        try:
            ct_validator.validate(template)
        except jsonschema.ValidationError as e:
            LOG.warning("Validation for {path} failed, {reason}".format(
                path=cl["path"], reason=e))
            return True
    return False


def add_config_section(section_name, options):
    if section_name and hasattr(CONF, section_name):
        # It's already been added
        return

    if section_name:
        group = cfg.OptGroup(name=section_name)
        CONF.register_group(group)
        CONF.register_opts(options, group)
    else:
        # Add options to the default section
        CONF.register_opts(options)


def add_config_section_for_template(template):
    '''Register a config section based on the template values

    Check to see if the configuration files contain a section
    that corresponds to the template.  If an appropriate section
    can be found, register options for the template so that the
    config values can be read and applied to the template via
    substitution (oslo supports registering groups and options
    at any time, before or after the config files are parsed).

    Corresponding section names may be of the following forms:

    <template_name>, example "hdp-2.0.6-master"
    This is useful when a template naming convention is being used,
    so that the template name is already unambiguous

    <plugin_name>_<hadoop_version>_<template_name>, example "hdp_2.0.6_master"
    This can be used if there is a name collision between templates

    <plugin_name>_<hadoop_version>, example "hdp_2.0.6"
    <plugin_name>, example "hdp"
    DEFAULT

    Sections are tried in the order given above.

    Since the first two section naming forms refer to a specific
    template by name, options are added based on template type.

    However, the other section naming forms may map to node group templates
    or cluster templates, so options for both are added.
    '''
    sections = list(CONF.list_all_sections())

    unique_name = "{name}".format(**template)
    fullname = "{plugin_name}_{hadoop_version}_{name}".format(**template)
    plugin_version = "{plugin_name}_{hadoop_version}".format(**template)
    plugin = "{plugin_name}".format(**template)

    section_name = None
    if unique_name in sections:
        section_name = unique_name
    elif fullname in sections:
        section_name = fullname

    if section_name:
        if u.is_node_group(template):
            opts = node_group_template_opts
        else:
            opts = cluster_template_opts
    else:
        if plugin_version in sections:
            section_name = plugin_version
        elif plugin in sections:
            section_name = plugin
        opts = all_template_opts

    add_config_section(section_name, opts)
    return section_name


def substitute_config_values(configs, template, path):
    if u.is_node_group(template):
        opt_names = node_group_template_opt_names
    else:
        opt_names = cluster_template_opt_names

    for opt, value in six.iteritems(configs):
        if opt in opt_names and opt in template:
            template[opt] = value


def get_configs(section):
    if section is None:
        return dict(CONF)
    return dict(CONF[section])


def get_plugin_name():
    if CONF.command.name == "update" and (
            not CONF.command.plugin_name and (
                hasattr(CONF, "plugins") and CONF.plugins)):
        return CONF.plugins
    return CONF.command.plugin_name


def process_files(dirname, files):

    node_groups = []
    clusters = []
    plugin_name = get_plugin_name()

    try:
        for fname in files:
            if os.path.splitext(fname)[1] == ".json":
                fpath = os.path.join(dirname, fname)
                with open(fpath, 'r') as fp:
                    try:
                        data = fp.read()
                        template = json.loads(data)
                    except ValueError as e:
                        LOG.warning("Error processing {path}, "
                                    "{reason}".format(path=fpath, reason=e))
                        raise Handled("error processing files")

                    # If this file doesn't contain basic fields, skip it.
                    # If we are filtering on plugin and version make
                    # sure the file is one that we want
                    if not u.check_basic_fields(template) or (
                            not u.check_plugin_name_and_version(
                                template,
                                plugin_name,
                                CONF.command.plugin_version)):
                        continue

                    # Look through the sections in CONF and register
                    # options for this template if we find a section
                    # related to the template (ie, plugin, version, name)
                    section = add_config_section_for_template(template)
                    LOG.debug("Using config section {section} "
                              "for {path}".format(section=section, path=fpath))

                    # Attempt to resolve substitutions using the config section
                    substitute_config_values(get_configs(section),
                                             template, fpath)

                    file_entry = {'template': template,
                                  'path': fpath}

                    if u.is_node_group(template):
                        # JSON validator
                        try:
                            ng_validator.validate(template)
                        except jsonschema.ValidationError as e:
                            LOG.warning("Validation for {path} failed, "
                                        "{reason}".format(path=fpath,
                                                          reason=e))
                            raise Handled(
                                "node group template validation failed")
                        node_groups.append(file_entry)
                        LOG.debug("Added {path} to node group "
                                  "template files".format(path=fpath))
                    else:
                        clusters.append(file_entry)
                        LOG.debug("Added {path} to cluster template "
                                  "files".format(path=fpath))

    except Handled as e:
        log_skipping_dir(dirname, str(e))
        node_groups = []
        clusters = []

    except Exception as e:
        log_skipping_dir(dirname,
                         "unhandled exception, {reason}".format(reason=e))
        node_groups = []
        clusters = []

    return node_groups, clusters


def delete_node_group_template(ctx, template, rollback=False):
    rollback_msg = " on rollback" if rollback else ""

    # If we are not deleting something that we just created,
    # do usage checks to ensure that the template is not in
    # use by a cluster or a cluster template
    if not rollback:
        clusters = conductor.API.cluster_get_all(ctx)
        cluster_templates = conductor.API.cluster_template_get_all(ctx)
        cluster_users, template_users = u.check_node_group_template_usage(
            template["id"], clusters, cluster_templates)

        if cluster_users:
            LOG.warning("Node group template {info} "
                        "in use by clusters {clusters}".format(
                            info=u.name_and_id(template),
                            clusters=cluster_users))
        if template_users:
            LOG.warning("Node group template {info} "
                        "in use by cluster templates {templates}".format(
                            info=u.name_and_id(template),
                            templates=template_users))

        if cluster_users or template_users:
            LOG.warning("Deletion of node group template "
                        "{info} failed".format(info=u.name_and_id(template)))
            return

    try:
        conductor.API.node_group_template_destroy(ctx, template["id"],
                                                  ignore_prot_on_def=True)
    except Exception as e:
        LOG.warning("Deletion of node group template {info} "
                    "failed{rollback}, {reason}".format(
                        info=u.name_and_id(template),
                        reason=e,
                        rollback=rollback_msg))
    else:
        LOG.info("Deleted node group template {info}{rollback}".format(
            info=u.name_and_id(template), rollback=rollback_msg))


def reverse_node_group_template_creates(ctx, templates):
    for template in templates:
        delete_node_group_template(ctx, template, rollback=True)


def reverse_node_group_template_updates(ctx, update_info):
    for template, values in update_info:
        # values are the original values that we overwrote in the update
        try:
            conductor.API.node_group_template_update(ctx,
                                                     template["id"], values,
                                                     ignore_prot_on_def=True)
        except Exception as e:
            LOG.warning("Rollback of update for node group "
                        "template {info} failed, {reason}".format(
                            info=u.name_and_id(template),
                            reason=e))
        else:
            LOG.info("Rolled back update for node group "
                     "template {info}".format(info=u.name_and_id(template)))


def add_node_group_templates(ctx, node_groups):

    error = False
    ng_info = {"ids": {},
               "created": [],
               "updated": []}

    def do_reversals(ng_info):
        reverse_node_group_template_updates(ctx, ng_info["updated"])
        reverse_node_group_template_creates(ctx, ng_info["created"])
        return {}, True

    try:
        for ng in node_groups:
            template = ng['template']
            current = u.find_node_group_template_by_name(ctx, template['name'])
            if current:

                # Track what we see in the current template that is different
                # from our update values. Save it for possible rollback.
                # Note, this is not perfect because it does not recurse through
                # nested structures to get an exact diff, but it ensures that
                # we track only fields that are valid in the JSON schema
                updated_fields = u.value_diff(current.to_dict(), template)

                # Always attempt to update.  Since the template value is a
                # combination of JSON and config values, there is no useful
                # timestamp we can use to skip an update.
                # If sqlalchemy determines no change in fields, it will not
                # mark it as updated.
                try:
                    template = conductor.API.node_group_template_update(
                        ctx, current['id'], template, ignore_prot_on_def=True)
                except Exception as e:
                    LOG.warning("Update of node group template {info} "
                                "failed, {reason}".format(
                                    info=u.name_and_id(current),
                                    reason=e))
                    raise Handled()

                if template['updated_at'] != current['updated_at']:
                    ng_info["updated"].append((template, updated_fields))
                    LOG.info("Updated node group template {info} "
                             "from {path}".format(info=u.name_and_id(template),
                                                  path=ng["path"]))
                else:
                    LOG.debug("No change to node group template {info} "
                              "from {path}".format(
                                  info=u.name_and_id(current),
                                  path=ng['path']))
            else:
                template['is_default'] = True
                try:
                    template = conductor.API.node_group_template_create(
                        ctx, template)
                except Exception as e:
                    LOG.warning("Creation of node group template "
                                "from {path} failed, {reason}".format(
                                    path=ng['path'], reason=e))
                    raise Handled()

                ng_info["created"].append(template)
                LOG.info("Created node group template {info} "
                         "from {path}".format(info=u.name_and_id(template),
                                              path=ng["path"]))

            # For the purposes of substitution we need a dict of id by name
            ng_info["ids"][template['name']] = template['id']

    except Handled:
        ng_info, error = do_reversals(ng_info)

    except Exception as e:
        LOG.warning("Unhandled exception while processing "
                    "node group templates, {reason}".format(reason=e))
        ng_info, error = do_reversals(ng_info)

    return ng_info, error


def delete_cluster_template(ctx, template, rollback=False):
    rollback_msg = " on rollback" if rollback else ""

    # If we are not deleting something that we just created,
    # do usage checks to ensure that the template is not in
    # use by a cluster
    if not rollback:
        clusters = conductor.API.cluster_get_all(ctx)
        cluster_users = u.check_cluster_template_usage(template["id"],
                                                       clusters)

        if cluster_users:
            LOG.warning("Cluster template {info} "
                        "in use by clusters {clusters}".format(
                            info=u.name_and_id(template),
                            clusters=cluster_users))

            LOG.warning("Deletion of cluster template "
                        "{info} failed".format(info=u.name_and_id(template)))
            return

    try:
        conductor.API.cluster_template_destroy(ctx, template["id"],
                                               ignore_prot_on_def=True)
    except Exception as e:
        LOG.warning("Deletion of cluster template {info} failed{rollback}"
                    ", {reason}".format(info=u.name_and_id(template),
                                        reason=e,
                                        rollback=rollback_msg))
    else:
        LOG.info("Deleted cluster template {info}{rollback}".format(
            info=u.name_and_id(template), rollback=rollback_msg))


def reverse_cluster_template_creates(ctx, templates):
    for template in templates:
        delete_cluster_template(ctx, template, rollback=True)


def reverse_cluster_template_updates(ctx, update_info):
    for template, values in update_info:
        # values are the original values that we overwrote in the update
        try:
            conductor.API.cluster_template_update(ctx,
                                                  template["id"], values,
                                                  ignore_prot_on_def=True)
        except Exception as e:
            LOG.warning("Rollback of update for cluster "
                        "template {info} failed, {reason}".format(
                            info=u.name_and_id(template),
                            reason=e))
        else:
            LOG.info("Rolled back update for cluster "
                     "template {info}".format(info=u.name_and_id(template)))


def add_cluster_templates(ctx, clusters, ng_dict):
    '''Add cluster templates to the database.

    The value of any node_group_template_id fields in cluster
    templates which reference a node group template in ng_dict by name
    will be changed to the id of the node group template.

    If there is an error in creating or updating a template, any templates
    that have already been created will be delete and any updates will
    be reversed.

    :param clusters: a list of dictionaries. Each dictionary
                     has a "template" entry holding the cluster template
                     and a "path" entry holding the path of the file
                     from which the template was read.
    :param ng_dict: a dictionary of node group template ids keyed
                    by node group template names
    '''

    error = False
    created = []
    updated = []

    def do_reversals(created, updated):
        reverse_cluster_template_updates(ctx, updated)
        reverse_cluster_template_creates(ctx, created)
        return True

    try:
        for cl in clusters:
            template = cl['template']

            # Fix up node_group_template_id fields
            u.substitute_ng_ids(template, ng_dict)

            # Name + tenant_id is unique, so search by name
            current = u.find_cluster_template_by_name(ctx, template['name'])
            if current:

                # Track what we see in the current template that is different
                # from our update values. Save it for possible rollback.
                # Note, this is not perfect because it does not recurse through
                # nested structures to get an exact diff, but it ensures that
                # we track only fields that are valid in the JSON schema
                updated_fields = u.value_diff(current.to_dict(), template)

                # Always attempt to update.  Since the template value is a
                # combination of JSON and config values, there is no useful
                # timestamp we can use to skip an update.
                # If sqlalchemy determines no change in fields, it will not
                # mark it as updated.

                # TODO(tmckay): why when I change the count in an
                # entry in node_groups does it not count as an update?
                # Probably a bug
                try:
                    template = conductor.API.cluster_template_update(
                        ctx, current['id'], template, ignore_prot_on_def=True)
                except Exception as e:
                    LOG.warning("Update of cluster template {info} "
                                "failed, {reason}".format(
                                    info=u.name_and_id(current), reason=e))
                    raise Handled()

                if template['updated_at'] != current['updated_at']:
                    updated.append((template, updated_fields))
                    LOG.info("Updated cluster template {info} "
                             "from {path}".format(info=u.name_and_id(template),
                                                  path=cl['path']))
                else:
                    LOG.debug("No change to cluster template {info} "
                              "from {path}".format(info=u.name_and_id(current),
                                                   path=cl["path"]))
            else:
                template["is_default"] = True
                try:
                    template = conductor.API.cluster_template_create(ctx,
                                                                     template)
                except Exception as e:
                    LOG.warning("Creation of cluster template "
                                "from {path} failed, {reason}".format(
                                    path=cl['path'],
                                    reason=e))
                    raise Handled()

                created.append(template)
                LOG.info("Created cluster template {info} "
                         "from {path}".format(info=u.name_and_id(template),
                                              path=cl['path']))

    except Handled:
        error = do_reversals(created, updated)

    except Exception as e:
        LOG.warning("Unhandled exception while processing "
                    "cluster templates, {reason}".format(reason=e))
        error = do_reversals(created, updated)

    return error


def do_update():
    '''Create or update default templates for the specified tenant.

    Looks for '.json' files beginning at the specified starting
    directory (--directory CLI option) and descending
    through subdirectories by default.

    The .json files represent cluster templates or node group
    templates. All '.json' files at the same directory level are treated
    as a set. Cluster templates may reference node group templates
    in the same set.

    If an error occurs in processing a set, skip it and continue.

    If creation of cluster templates fails, any node group templates
    in the set that were already created will be deleted.
    '''

    ctx = Context(tenant_id=CONF.command.tenant_id)
    start_dir = os.path.abspath(CONF.command.directory)

    for root, dirs, files in os.walk(start_dir):

        # Find all the template files and identify them as node_group
        # or cluster templates. If there is an exception in
        # processing the set, return empty lists.
        node_groups, clusters = process_files(root, files)

        # Now that we know what the valid node group templates are,
        # we can validate the cluster templates as well.
        if check_cluster_templates_valid(node_groups, clusters):
            log_skipping_dir(root, "error processing cluster templates")

        # If there are existing default templates that match the names
        # in the template files, do usage checks here to detect update
        # failures early (we can't update a template in use)
        elif check_usage_of_existing(ctx, node_groups, clusters):
            log_skipping_dir(root, "templates in use")
        else:
            ng_info, error = add_node_group_templates(ctx, node_groups)
            if error:
                log_skipping_dir(root, "error processing node group templates")

            elif add_cluster_templates(ctx, clusters, ng_info["ids"]):
                log_skipping_dir(root, "error processing cluster templates")

                # Cluster templates failed so remove the node group templates
                reverse_node_group_template_updates(ctx, ng_info["updated"])
                reverse_node_group_template_creates(ctx, ng_info["created"])

        if CONF.command.norecurse:
            break


def do_delete():
    '''Delete default templates in the specified tenant

    Deletion uses the --plugin-name and --plugin-version options
    as filters.

    Only templates with 'is_default=True' will be deleted.
    '''

    ctx = Context(tenant_id=CONF.command.tenant_id)

    for plugin in get_plugin_name():

        kwargs = {'is_default': True}
        kwargs['plugin_name'] = plugin

        # Delete cluster templates first for the sake of usage checks
        lst = conductor.API.cluster_template_get_all(ctx, **kwargs)
        for l in lst:
            if not u.check_plugin_version(l, CONF.command.plugin_version):
                continue
            delete_cluster_template(ctx, l)

        lst = conductor.API.node_group_template_get_all(ctx, **kwargs)
        for l in lst:
            if not u.check_plugin_version(l, CONF.command.plugin_version):
                continue
            delete_node_group_template(ctx, l)


def do_node_group_template_delete():
    ctx = Context(tenant_id=CONF.command.tenant_id)

    template_name = CONF.command.template_name
    t = u.find_node_group_template_by_name(ctx, template_name)
    if t:
        delete_node_group_template(ctx, t)
    else:
        LOG.warning("Deletion of node group template {name} failed, "
                    "no such template".format(name=template_name))


def do_node_group_template_delete_by_id():
    ctx = Context(is_admin=True)

    # Make sure it's a default
    t = conductor.API.node_group_template_get(ctx, CONF.command.id)
    if t:
        if t["is_default"]:
            delete_node_group_template(ctx, t)
        else:
            LOG.warning("Deletion of node group template {info} skipped, "
                        "not a default template".format(
                            info=u.name_and_id(t)))
    else:
        LOG.warning("Deletion of node group template {id} failed, "
                    "no such template".format(id=CONF.command.id))


def do_cluster_template_delete():
    ctx = Context(tenant_id=CONF.command.tenant_id)

    template_name = CONF.command.template_name
    t = u.find_cluster_template_by_name(ctx, template_name)
    if t:
        delete_cluster_template(ctx, t)
    else:
        LOG.warning("Deletion of cluster template {name} failed, "
                    "no such template".format(name=template_name))


def do_cluster_template_delete_by_id():
    ctx = Context(is_admin=True)

    # Make sure it's a default
    t = conductor.API.cluster_template_get(ctx, CONF.command.id)
    if t:
        if t["is_default"]:
            delete_cluster_template(ctx, t)
        else:
            LOG.warning("Deletion of cluster template {info} skipped, "
                        "not a default template".format(
                            info=u.name_and_id(t)))
    else:
        LOG.warning("Deletion of cluster template {id} failed, "
                    "no such template".format(id=CONF.command.id))



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add_job_interface

Revision ID: 022
Revises: 021
Create Date: 2015-01-27 15:53:22.128263

"""

# revision identifiers, used by Alembic.
revision = '022'
down_revision = '021'

from alembic import op
import sqlalchemy as sa

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    op.create_table('job_interface_arguments',
                    sa.Column('created_at', sa.DateTime(),
                              nullable=True),
                    sa.Column('updated_at', sa.DateTime(),
                              nullable=True),
                    sa.Column('id', sa.String(length=36),
                              nullable=False),
                    sa.Column('job_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('tenant_id', sa.String(length=36),
                              nullable=True),
                    sa.Column('name', sa.String(80),
                              nullable=False),
                    sa.Column('description', sa.Text()),
                    sa.Column('mapping_type', sa.String(80),
                              nullable=False),
                    sa.Column('location', sa.Text(),
                              nullable=False),
                    sa.Column('value_type', sa.String(80),
                              nullable=False),
                    sa.Column('required', sa.Boolean(),
                              nullable=False),
                    sa.Column('order', sa.SmallInteger(),
                              nullable=False),
                    sa.Column('default', sa.Text()),
                    sa.ForeignKeyConstraint(['job_id'],
                                            ['jobs.id']),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('job_id', 'order'),
                    sa.UniqueConstraint('job_id', 'name'),
                    mysql_engine=MYSQL_ENGINE,
                    mysql_charset=MYSQL_CHARSET)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sahara.exceptions as e
from sahara.i18n import _
from sahara.service.api import v11 as api
from sahara.service.validations.edp import job_interface as j_i
from sahara.utils import edp


def _check_binaries(values):
    for job_binary in values:
        if not api.get_job_binary(job_binary):
            raise e.NotFoundException(job_binary,
                                      _("Job binary '%s' does not exist"))


def check_mains_libs(data, **kwargs):
    mains = data.get("mains", [])
    libs = data.get("libs", [])
    job_type, subtype = edp.split_job_type(data.get("type"))
    streaming = (job_type == edp.JOB_TYPE_MAPREDUCE and
                 subtype == edp.JOB_SUBTYPE_STREAMING)

    # These types must have a value in mains and may also use libs
    if job_type in [edp.JOB_TYPE_PIG, edp.JOB_TYPE_HIVE,
                    edp.JOB_TYPE_SHELL, edp.JOB_TYPE_SPARK,
                    edp.JOB_TYPE_STORM, edp.JOB_TYPE_PYLEUS]:
        if not mains:
            if job_type in [edp.JOB_TYPE_SPARK, edp.JOB_TYPE_STORM,
                            edp.JOB_TYPE_PYLEUS]:
                msg = _(
                    "%s job requires main application jar") % data.get("type")
            else:
                msg = _("%s flow requires main script") % data.get("type")
            raise e.InvalidDataException(msg)

        # Check for overlap
        if set(mains).intersection(set(libs)):
            raise e.InvalidDataException(_("'mains' and 'libs' overlap"))

    else:
        # Java and MapReduce require libs, but MapReduce.Streaming does not
        if not streaming and not libs:
            raise e.InvalidDataException(_("%s flow requires libs") %
                                         data.get("type"))
        if mains:
            raise e.InvalidDataException(_("%s flow does not use mains") %
                                         data.get("type"))

    # Make sure that all referenced binaries exist
    _check_binaries(mains)
    _check_binaries(libs)


def check_interface(data, **kwargs):
    j_i.check_job_interface(data, **kwargs)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.swift import utils
from sahara.tests.unit import base as testbase


class SwiftUtilsTest(testbase.SaharaTestCase):

    def setUp(self):
        super(SwiftUtilsTest, self).setUp()
        self.override_config('use_identity_api_v3', True)
        self.setup_context(service_catalog=True)

    @mock.patch('sahara.utils.openstack.base.url_for')
    def test_retrieve_auth_url(self, url_for_mock):
        correct = "https://127.0.0.1:8080/v3/"

        def _assert(uri):
            url_for_mock.return_value = uri
            self.assertEqual(correct, utils.retrieve_auth_url())

        _assert("%s/" % correct)
        _assert("https://127.0.0.1:8080")
        _assert("https://127.0.0.1:8080/")
        _assert("https://127.0.0.1:8080/v3")
        _assert("https://127.0.0.1:8080/v3/")

    @mock.patch('sahara.utils.openstack.base.url_for')
    def test_retrieve_auth_url_path_present(self, url_for_mock):
        correct = "https://127.0.0.1:8080/identity/v3/"

        def _assert(uri):
            url_for_mock.return_value = uri
            self.assertEqual(correct, utils.retrieve_auth_url())

        _assert("https://127.0.0.1:8080/identity")
        _assert("https://127.0.0.1:8080/identity/v3/")

    @mock.patch('sahara.utils.openstack.base.url_for')
    def test_retrieve_auth_url_without_port(self, url_for_mock):
        correct = "https://127.0.0.1/v3/"

        def _assert(uri):
            url_for_mock.return_value = uri
            self.assertEqual(correct, utils.retrieve_auth_url())

        _assert("%s/" % correct)
        _assert("https://127.0.0.1")
        _assert("https://127.0.0.1/")
        _assert("https://127.0.0.1/v3")
        _assert("https://127.0.0.1/v3/")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

import sahara.exceptions as ex
from sahara.service.edp.oozie.workflow_creator import hive_workflow as hw
from sahara.service.edp.oozie.workflow_creator import java_workflow as jw
from sahara.service.edp.oozie.workflow_creator import mapreduce_workflow as mrw
from sahara.service.edp.oozie.workflow_creator import pig_workflow as pw
from sahara.service.edp.oozie.workflow_creator import shell_workflow as shw
from sahara.utils import patches as p


class TestWorkflowCreators(testtools.TestCase):

    def setUp(self):
        super(TestWorkflowCreators, self).setUp()
        p.patch_minidom_writexml()
        self.prepare = {'delete': ['delete_dir_1', 'delete_dir_2'],
                        'mkdir': ['mkdir_1']}
        self.job_xml = 'job_xml.xml'
        self.configuration = {'conf_param_1': 'conf_value_1',
                              'conf_param_2': 'conf_value_3'}
        self.files = ['file1', 'file2']
        self.archives = ['arch1']
        self.streaming = {'mapper': '/usr/bin/cat',
                          'reducer': '/usr/bin/wc'}

    def test_create_mapreduce_streaming(self):
        mr_action = """
      <streaming>
        <mapper>/usr/bin/cat</mapper>
        <reducer>/usr/bin/wc</reducer>
      </streaming>"""

        mr_workflow = mrw.MapReduceWorkFlowCreator()
        mr_workflow.build_workflow_xml(self.prepare, self.job_xml,
                                       self.configuration, self.files,
                                       self.archives, self.streaming)
        res = mr_workflow.get_built_workflow_xml()
        self.assertIn(mr_action, res)

        mr_workflow = mrw.MapReduceWorkFlowCreator()
        mr_workflow.build_workflow_xml(self.prepare, self.job_xml,
                                       self.configuration, self.files,
                                       self.archives)
        res = mr_workflow.get_built_workflow_xml()
        self.assertNotIn(mr_action, res)

        mr_workflow = mrw.MapReduceWorkFlowCreator()
        with testtools.ExpectedException(ex.NotFoundException):
            mr_workflow.build_workflow_xml(self.prepare, self.job_xml,
                                           self.configuration, self.files,
                                           self.archives, {'bogus': 'element'})

    def test_create_mapreduce_workflow(self):
        mr_workflow = mrw.MapReduceWorkFlowCreator()
        mr_workflow.build_workflow_xml(self.prepare, self.job_xml,
                                       self.configuration, self.files,
                                       self.archives)
        res = mr_workflow.get_built_workflow_xml()
        mr_action = """    <map-reduce>
      <job-tracker>${jobTracker}</job-tracker>
      <name-node>${nameNode}</name-node>
      <prepare>
        <delete path="delete_dir_1"/>
        <delete path="delete_dir_2"/>
        <mkdir path="mkdir_1"/>
      </prepare>
      <job-xml>job_xml.xml</job-xml>
      <configuration>
        <property>
          <name>conf_param_1</name>
          <value>conf_value_1</value>
        </property>
        <property>
          <name>conf_param_2</name>
          <value>conf_value_3</value>
        </property>
      </configuration>
      <file>file1</file>
      <file>file2</file>
      <archive>arch1</archive>
    </map-reduce>"""

        self.assertIn(mr_action, res)

    def test_create_pig_workflow(self):
        pig_workflow = pw.PigWorkflowCreator()

        pig_script = 'script.pig'
        param_dict = {'param1': 'param_value1'}
        args = ['arg_value1', 'arg_value2']

        pig_workflow.build_workflow_xml(pig_script, self.prepare,
                                        self.job_xml, self.configuration,
                                        param_dict, args,
                                        self.files, self.archives)
        res = pig_workflow.get_built_workflow_xml()
        pig_action = """    <pig>
      <job-tracker>${jobTracker}</job-tracker>
      <name-node>${nameNode}</name-node>
      <prepare>
        <delete path="delete_dir_1"/>
        <delete path="delete_dir_2"/>
        <mkdir path="mkdir_1"/>
      </prepare>
      <job-xml>job_xml.xml</job-xml>
      <configuration>
        <property>
          <name>conf_param_1</name>
          <value>conf_value_1</value>
        </property>
        <property>
          <name>conf_param_2</name>
          <value>conf_value_3</value>
        </property>
      </configuration>
      <script>script.pig</script>
      <param>param1=param_value1</param>
      <argument>arg_value1</argument>
      <argument>arg_value2</argument>
      <file>file1</file>
      <file>file2</file>
      <archive>arch1</archive>
    </pig>"""

        self.assertIn(pig_action, res)

    def test_create_hive_workflow(self):
        hive_workflow = hw.HiveWorkflowCreator()
        hive_script = "script.q"
        params = {"key": "value", "key2": "value2"}
        hive_workflow.build_workflow_xml(hive_script, self.job_xml,
                                         self.prepare, self.configuration,
                                         params, self.files, self.archives)
        res = hive_workflow.get_built_workflow_xml()
        hive_action = """  <hive xmlns="uri:oozie:hive-action:0.2">
      <job-tracker>${jobTracker}</job-tracker>
      <name-node>${nameNode}</name-node>
      <prepare>
        <delete path="delete_dir_1"/>
        <delete path="delete_dir_2"/>
        <mkdir path="mkdir_1"/>
      </prepare>
      <job-xml>job_xml.xml</job-xml>
      <configuration>
        <property>
          <name>conf_param_1</name>
          <value>conf_value_1</value>
        </property>
        <property>
          <name>conf_param_2</name>
          <value>conf_value_3</value>
        </property>
      </configuration>
      <script>script.q</script>
      <param>key=value</param>
      <param>key2=value2</param>
      <file>file1</file>
      <file>file2</file>
      <archive>arch1</archive>
    </hive>"""

        self.assertIn(hive_action, res)

    def test_create_java_workflow(self):
        java_workflow = jw.JavaWorkflowCreator()
        main_class = 'org.apache.hadoop.examples.SomeClass'
        args = ['/user/hadoop/input',
                '/user/hadoop/output']
        java_opts = '-Dparam1=val1 -Dparam2=val2'

        java_workflow.build_workflow_xml(main_class,
                                         self.prepare,
                                         self.job_xml, self.configuration,
                                         java_opts, args,
                                         self.files, self.archives)
        res = java_workflow.get_built_workflow_xml()
        java_action = """
      <job-tracker>${jobTracker}</job-tracker>
      <name-node>${nameNode}</name-node>
      <prepare>
        <delete path="delete_dir_1"/>
        <delete path="delete_dir_2"/>
        <mkdir path="mkdir_1"/>
      </prepare>
      <job-xml>job_xml.xml</job-xml>
      <configuration>
        <property>
          <name>conf_param_1</name>
          <value>conf_value_1</value>
        </property>
        <property>
          <name>conf_param_2</name>
          <value>conf_value_3</value>
        </property>
      </configuration>
      <main-class>org.apache.hadoop.examples.SomeClass</main-class>
      <java-opts>-Dparam1=val1 -Dparam2=val2</java-opts>
      <arg>/user/hadoop/input</arg>
      <arg>/user/hadoop/output</arg>
      <file>file1</file>
      <file>file2</file>
      <archive>arch1</archive>
    </java>"""

        self.assertIn(java_action, res)

    def test_create_shell_workflow(self):
        shell_workflow = shw.ShellWorkflowCreator()
        main_class = 'doit.sh'
        args = ['now']
        env_vars = {"VERSION": 3}

        shell_workflow.build_workflow_xml(main_class,
                                          self.prepare,
                                          self.job_xml,
                                          self.configuration,
                                          env_vars,
                                          args,
                                          self.files)

        res = shell_workflow.get_built_workflow_xml()
        shell_action = """
    <shell xmlns="uri:oozie:shell-action:0.1">
      <job-tracker>${jobTracker}</job-tracker>
      <name-node>${nameNode}</name-node>
      <prepare>
        <delete path="delete_dir_1"/>
        <delete path="delete_dir_2"/>
        <mkdir path="mkdir_1"/>
      </prepare>
      <configuration>
        <property>
          <name>conf_param_1</name>
          <value>conf_value_1</value>
        </property>
        <property>
          <name>conf_param_2</name>
          <value>conf_value_3</value>
        </property>
      </configuration>
      <exec>doit.sh</exec>
      <argument>now</argument>
      <env-var>VERSION=3</env-var>
      <file>file1</file>
      <file>file2</file>
      <file>doit.sh</file>
    </shell>"""

        self.assertIn(shell_action, res)



# Copyright (c) 2017 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from sahara.utils.openstack import heat as heat_u


class HeatClientTest(testtools.TestCase):
    @mock.patch('sahara.utils.openstack.heat.get_stack')
    @mock.patch('heatclient.client.Client')
    @mock.patch('sahara.utils.openstack.base.url_for')
    @mock.patch('sahara.service.sessions.cache')
    @mock.patch('sahara.context.ctx')
    def test_deleting(self, ctx, cache, url_for, heat, get_stack):
        class _FakeHeatStacks(object):
            def delete(self, stack):
                call_list.append("delete")

        call_list = None
        get_stack.return_value = None
        get_stack.side_effect = lambda *args, **kwargs: call_list.append("get")

        heat.return_value.stacks = _FakeHeatStacks()

        call_list = []
        heat_u.lazy_delete_stack(mock.Mock())
        self.assertEqual(call_list, ["delete"])

        call_list = []
        heat_u.delete_stack(mock.Mock())
        self.assertEqual(call_list, ["delete", "get"])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara import conductor as c


conductor = c.API


def get_raw_data(context, job_binary):
    # url example: 'internal-db://JobBinaryInternal-UUID'
    binary_internal_id = job_binary.url[len("internal-db://"):]
    return conductor.job_binary_internal_get_raw_data(context,
                                                      binary_internal_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.service.api import v10 as api
from sahara.service.validations.edp import job_binary as b
from sahara.service.validations.edp.job_binary import jb_manager
from sahara.service.validations.edp import job_binary_schema as b_s
from sahara.swift import utils as su
from sahara.tests.unit.service.validation import utils as u


class TestJobBinaryValidation(u.ValidationTestCase):
    def setUp(self):
        super(TestJobBinaryValidation, self).setUp()
        self._create_object_fun = b.check_job_binary
        self.scheme = b_s.JOB_BINARY_SCHEMA
        api.plugin_base.setup_plugins()
        jb_manager.setup_job_binaries()

    @mock.patch('sahara.utils.api_validator.jb_manager')
    def test_creation(self, mock_jb_manager):
        JOB_BINARIES = mock.Mock()
        mock_jb = mock.Mock()
        mock_jb_manager.JOB_BINARIES = JOB_BINARIES

        JOB_BINARIES.get_job_binary_by_url = mock.Mock(return_value=mock_jb)

        mock_jb.validate_job_location_format = mock.Mock(return_value=True)

        data = {
            "name": "main.jar",
            "url": "internal-db://3e4651a5-1f08-4880-94c4-596372b37c64",
            "extra": {
                "user": "user",
                "password": "password"
            },
            "description": "long description"
        }
        self._assert_types(data)

    @mock.patch('sahara.utils.api_validator.jb_manager')
    def test_job_binary_create_swift(self, mock_jb_manager):
        JOB_BINARIES = mock.Mock()
        mock_jb = mock.Mock()
        mock_jb_manager.JOB_BINARIES = JOB_BINARIES

        JOB_BINARIES.get_job_binary_by_url = mock.Mock(return_value=mock_jb)

        mock_jb.validate_job_location_format = mock.Mock(return_value=True)
        self._assert_create_object_validation(
            data={
                "name": "j_o_w",
                "url": su.SWIFT_INTERNAL_PREFIX + "o.sahara/k"
            },
            bad_req_i=(1, "BAD_JOB_BINARY",
                       "To work with JobBinary located in internal "
                       "swift add 'user' and 'password' to extra"))
        self.override_config('use_domain_for_proxy_users', True)
        self._assert_create_object_validation(
            data={
                "name": "j_o_w",
                "url": su.SWIFT_INTERNAL_PREFIX + "o.sahara/k"
            })

    @mock.patch('sahara.utils.api_validator.jb_manager')
    def test_job_binary_create_internal(self, mock_jb_manager):
        JOB_BINARIES = mock.Mock()
        mock_jb = mock.Mock()
        mock_jb_manager.JOB_BINARIES = JOB_BINARIES

        JOB_BINARIES.get_job_binary_by_url = mock.Mock(return_value=mock_jb)

        mock_jb.validate_job_location_format = mock.Mock(return_value=False)
        self._assert_create_object_validation(
            data={
                "name": "main.jar",
                "url": "internal-db://abacaba",
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "url: 'internal-db://abacaba' is not a "
                       "'valid_job_location'"))

    @mock.patch('sahara.utils.api_validator.jb_manager')
    def test_job_binary_create_manila(self, mock_jb_manager):
        JOB_BINARIES = mock.Mock()
        mock_jb = mock.Mock()
        mock_jb_manager.JOB_BINARIES = JOB_BINARIES

        JOB_BINARIES.get_job_binary_by_url = mock.Mock(return_value=mock_jb)

        mock_jb.validate_job_location_format = mock.Mock(return_value=False)
        self._assert_create_object_validation(
            data={
                "name": "main.jar",
                "url": "manila://abacaba",
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       "url: 'manila://abacaba' is not a "
                       "'valid_job_location'"))
